# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Helpers for intersecting B |eacute| zier curves via algebraic methods.

Primarily helps implicitize B |eacute| zier curves.

.. _resultant: https://en.wikipedia.org/wiki/Resultant
.. _algebraic curve: https://en.wikipedia.org/wiki/Algebraic_curve
.. _Farouki and Rajan: https://dx.doi.org/10.1016/0167-8396(88)90016-7
.. _theorem: https://en.wikipedia.org/wiki/B%C3%A9zout's_theorem

Uses the `resultant`_ to evaluate the implicitized
`algebraic curve`_. In order to do this on B |eacute| zier curves
without translating to a power basis, we utilize the work of
`Farouki and Rajan`_ to compute a modified Sylvester determinant.

Given two parametric curves :math:`(x_1(s), y_1(s))` and
:math:`(x_2(t), y_2(t))`, we can determine an "intersection polynomial"
for both :math:`s` and :math:`t`. For example, by implicitizing the
first curve, we determine :math:`f_1(x, y)` and plugging the second
curve into this we find

.. math::

   g(t) = f_1\left(x_2(t), y_2(t)\right) = 0

is the "intersection polynomial" for :math:`t`.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import numpy as np
from numpy.polynomial import polynomial

try:
    import scipy.linalg.lapack as _scipy_lapack
except ImportError:  # pragma: NO COVER
    _scipy_lapack = None

from bezier import _curve_helpers
from bezier import _geometric_intersection
from bezier import _helpers
from bezier import _intersection_helpers
from bezier import _py_geometric_intersection
from bezier import _py_helpers


# NOTE: These are hardcoded from:
#         0.5 * (np.cos(np.pi * np.arange(1, 14, 2.0) / 14) + 1.0)
#         0.5 * (np.cos(np.pi * np.arange(1, 18, 2.0) / 18) + 1.0)
#         0.5 * (np.cos(np.pi * np.arange(1, 20, 2.0) / 20) + 1.0)
#       See 73f909805e971221cb7976cf69603b82f31a4a32 and
#       80907b1be03f5895f7132e0e920c5e3cdebba9ac.
_CHEB7 = np.asfortranarray(
    [
        float.fromhex("0x1.f994e02ac74b4p-1"),
        float.fromhex("0x1.c8261ba82ef26p-1"),
        float.fromhex("0x1.6f130135c6af0p-1"),
        float.fromhex("0x1.0000000000000p-1"),
        float.fromhex("0x1.21d9fd9472a20p-2"),
        float.fromhex("0x1.becf22be886e0p-4"),
        float.fromhex("0x1.9ac7f54e2d2c0p-7"),
    ]
)
_CHEB9 = np.asfortranarray(
    [
        float.fromhex("0x1.fc1c5c6408e0cp-1"),
        float.fromhex("0x1.ddb3d742c2656p-1"),
        float.fromhex("0x1.a48dba91e0b0ep-1"),
        float.fromhex("0x1.578ea1d2282fep-1"),
        float.fromhex("0x1.0000000000000p-1"),
        float.fromhex("0x1.50e2bc5bafa08p-2"),
        float.fromhex("0x1.6dc915b87d3c6p-3"),
        float.fromhex("0x1.126145e9ecd5cp-4"),
        float.fromhex("0x1.f1d1cdfb8fa40p-8"),
    ]
)
_CHEB10 = np.asfortranarray(
    [
        float.fromhex("0x1.fcd924a17f22ep-1"),
        float.fromhex("0x1.e41900e9e9636p-1"),
        float.fromhex("0x1.b504f333f9de6p-1"),
        float.fromhex("0x1.7438b8ad13780p-1"),
        float.fromhex("0x1.280c16cf50a6fp-1"),
        float.fromhex("0x1.afe7d2615eb25p-2"),
        float.fromhex("0x1.178e8ea5d9100p-2"),
        float.fromhex("0x1.2bec333018868p-3"),
        float.fromhex("0x1.be6ff16169ca0p-5"),
        float.fromhex("0x1.936daf406e940p-8"),
    ]
)
# Allow a buffer of sqrt(sqrt(machine precision)) for polynomial roots.
_IMAGINARY_WIGGLE = 0.5 ** 13
_UNIT_INTERVAL_WIGGLE_START = -(0.5 ** 13)
_UNIT_INTERVAL_WIGGLE_END = 1.0 + 0.5 ** 13
_SIGMA_THRESHOLD = 0.5 ** 20
_SINGULAR_EPS = 0.5 ** 52
# Detect almost zero polynomials.
_L2_THRESHOLD = 0.5 ** 40  # 4096 (machine precision)
_ZERO_THRESHOLD = 0.5 ** 38  # 16384 (machine precision)
_COEFFICIENT_THRESHOLD = 0.5 ** 26  # sqrt(machine precision)
_NON_SIMPLE_THRESHOLD = 0.5 ** 48  # 16 (machine precision)
_COINCIDENT_ERR = "Coincident curves not currently supported"
_NON_SIMPLE_ERR = "Polynomial has non-simple roots"
_POWER_BASIS_ERR = (
    "Currently only supporting degree pairs "
    "1-1, 1-2, 1-3, 1-4, 2-2, 2-3, 2-4 and 3-3."
)
_LINEARIZATION = _py_geometric_intersection.Linearization
_DISJOINT = _py_geometric_intersection.BoxIntersectionType.DISJOINT


def _evaluate3(nodes, x_val, y_val):
    """Helper for :func:`evaluate` when ``nodes`` is degree 3.

    Args:
        nodes (numpy.ndarray): ``2 x 4`` array of nodes in a curve.
        x_val (float): ``x``-coordinate for evaluation.
        y_val (float): ``y``-coordinate for evaluation.

    Returns:
        float: The computed value of :math:`f(x, y)`.
    """
    # NOTE: This may be (a) slower and (b) less precise than
    #       hard-coding the determinant.
    sylvester_mat = np.zeros((6, 6), order="F")
    delta = nodes - np.asfortranarray([[x_val], [y_val]])
    delta[:, 1:3] *= 3.0
    # Swap rows/columns so that x-y are right next to each other.
    # This will only change the determinant up to a sign.
    sylvester_mat[:2, :4] = delta
    sylvester_mat[2:4, 1:5] = delta
    sylvester_mat[4:, 2:] = delta
    return np.linalg.det(sylvester_mat)


def evaluate(nodes, x_val, y_val):
    r"""Evaluate the implicitized bivariate polynomial containing the curve.

    Assumes `algebraic curve`_ containing :math:`B(s, t)` is given by
    :math:`f(x, y) = 0`. This function evaluates :math:`f(x, y)`.

    .. note::

       This assumes, but doesn't check, that ``nodes`` has 2 rows.

    .. note::

       This assumes, but doesn't check, that ``nodes`` is not degree-elevated.
       If it were degree-elevated, then the Sylvester matrix will always
       have zero determinant.

    Args:
        nodes (numpy.ndarray): ``2 x N`` array of nodes in a curve.
        x_val (float): ``x``-coordinate for evaluation.
        y_val (float): ``y``-coordinate for evaluation.

    Returns:
        float: The computed value of :math:`f(x, y)`.

    Raises:
        ValueError: If the curve is a point.
        .UnsupportedDegree: If the degree is not 1, 2 or 3.
    """
    _, num_nodes = nodes.shape
    if num_nodes == 1:
        raise ValueError("A point cannot be implicitized")

    if num_nodes == 2:
        # x(s) - x = (x0 - x) (1 - s) + (x1 - x) s
        # y(s) - y = (y0 - y) (1 - s) + (y1 - y) s
        # Modified Sylvester: [x0 - x, x1 - x]
        #                     [y0 - y, y1 - y]
        return (nodes[0, 0] - x_val) * (nodes[1, 1] - y_val) - (
            nodes[0, 1] - x_val
        ) * (nodes[1, 0] - y_val)

    if num_nodes == 3:
        # x(s) - x = (x0 - x) (1 - s)^2 + 2 (x1 - x) s(1 - s) + (x2 - x) s^2
        # y(s) - y = (y0 - y) (1 - s)^2 + 2 (y1 - y) s(1 - s) + (y2 - y) s^2
        # Modified Sylvester: [x0 - x, 2(x1 - x),    x2 - x,      0] = A|B|C|0
        #                     [     0,    x0 - x, 2(x1 - x), x2 - x]   0|A|B|C
        #                     [y0 - y, 2(y1 - y),    y2 - y,      0]   D|E|F|0
        #                     [     0,    y0 - y, 2(y1 - y), y2 - y]   0|D|E|F
        val_a, val_b, val_c = nodes[0, :] - x_val
        val_b *= 2
        val_d, val_e, val_f = nodes[1, :] - y_val
        val_e *= 2
        #     [A, B, C]         [E, F, 0]
        # det [E, F, 0] = - det [A, B, C] = -E (BF - CE) + F(AF - CD)
        #     [D, E, F]         [D, E, F]
        sub1 = val_b * val_f - val_c * val_e
        sub2 = val_a * val_f - val_c * val_d
        sub_det_a = -val_e * sub1 + val_f * sub2
        #     [B, C, 0]
        # det [A, B, C] = B (BF - CE) - C (AF - CD)
        #     [D, E, F]
        sub_det_d = val_b * sub1 - val_c * sub2
        return val_a * sub_det_a + val_d * sub_det_d

    if num_nodes == 4:
        return _evaluate3(nodes, x_val, y_val)

    raise _py_helpers.UnsupportedDegree(num_nodes - 1, supported=(1, 2, 3))


def eval_intersection_polynomial(nodes1, nodes2, t):
    r"""Evaluates a parametric curve **on** an implicitized algebraic curve.

    Uses :func:`evaluate` to evaluate :math:`f_1(x, y)`, the implicitization
    of ``nodes1``. Then plugs ``t`` into the second parametric curve to
    get an ``x``- and ``y``-coordinate and evaluate the
    **intersection polynomial**:

    .. math::

       g(t) = f_1\left(x_2(t), y_2(t)right)

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.
        t (float): The parameter along ``nodes2`` where we evaluate
            the function.

    Returns:
        float: The computed value of :math:`f_1(x_2(t), y_2(t))`.
    """
    (x_val,), (y_val,) = _curve_helpers.evaluate_multi(
        nodes2, np.asfortranarray([t])
    )
    return evaluate(nodes1, x_val, y_val)


def _to_power_basis11(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that each curve is
    degree one. In this case, B |eacute| zout's `theorem`_ tells us
    that the **intersection polynomial** is degree :math:`1 \cdot 1`
    hence we return two coefficients.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``2``-array of coefficients.
    """
    # We manually invert the Vandermonde matrix:
    # [1 0][c0] = [n0]
    # [1 1][c1]   [n1]
    val0 = eval_intersection_polynomial(nodes1, nodes2, 0.0)
    val1 = eval_intersection_polynomial(nodes1, nodes2, 1.0)
    # [c0] = [ 1 0][n0]
    # [c1] = [-1 1][n1]
    return np.asfortranarray([val0, -val0 + val1])


def _to_power_basis12(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that the first curve is
    degree one and the second is degree two. In this case, B |eacute| zout's
    `theorem`_ tells us that the **intersection polynomial** is
    degree :math:`1 \cdot 2` hence we return three coefficients.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``3``-array of coefficients.
    """
    # We manually invert the Vandermonde matrix:
    # [1 0   0  ][c0] = [n0]
    # [1 1/2 1/4][c1]   [n1]
    # [1 1   1  ][c2]   [n2]
    val0 = eval_intersection_polynomial(nodes1, nodes2, 0.0)
    val1 = eval_intersection_polynomial(nodes1, nodes2, 0.5)
    val2 = eval_intersection_polynomial(nodes1, nodes2, 1.0)
    # [c0] = [ 1  0  0][n0]
    # [c1] = [-3  4 -1][n1]
    # [c2] = [ 2 -4  2][n2]
    return np.asfortranarray(
        [
            val0,
            -3.0 * val0 + 4.0 * val1 - val2,
            2.0 * val0 - 4.0 * val1 + 2.0 * val2,
        ]
    )


def _to_power_basis13(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that the first curve is
    degree one and the second is degree three. In this case, B |eacute| zout's
    `theorem`_ tells us that the **intersection polynomial** is
    degree :math:`1 \cdot 3` hence we return four coefficients.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``4``-array of coefficients.
    """
    # We manually invert the Vandermonde matrix:
    # Use exact f.p. numbers to avoid round-off wherever possible.
    # [1 0   0    0    ][c0] = [n0]
    # [1 1/4 1/16 1/64 ][c1]   [n1]
    # [1 3/4 9/16 27/64][c2]   [n2]
    # [1 1   1    1    ][c3]   [n3]
    val0 = eval_intersection_polynomial(nodes1, nodes2, 0.0)
    val1 = eval_intersection_polynomial(nodes1, nodes2, 0.25)
    val2 = eval_intersection_polynomial(nodes1, nodes2, 0.75)
    val3 = eval_intersection_polynomial(nodes1, nodes2, 1.0)
    # [c0] =       [  3   0   0   0][n0]
    # [c1] = 1 / 3 [-19  24  -8   3][n1]
    # [c2] =       [ 32 -56  40 -16][n2]
    # [c3] =       [-16  32 -32  16][n3]
    # Since polynomial coefficients, we don't need to divide by 3
    # to get the same polynomial. Avoid the division to avoid round-off.
    return np.asfortranarray(
        [
            3.0 * val0,
            -19.0 * val0 + 24.0 * val1 - 8.0 * val2 + 3.0 * val3,
            32.0 * val0 - 56.0 * val1 + 40.0 * val2 - 16.0 * val3,
            -16.0 * val0 + 32.0 * val1 - 32.0 * val2 + 16.0 * val3,
        ]
    )


def _to_power_basis_degree4(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that B |eacute| zout's
    `theorem`_ tells us the **intersection polynomial** is degree
    :math:`4`. This happens if the two curves have degrees two and two
    or have degrees one and four.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``5``-array of coefficients.
    """
    # We manually invert the Vandermonde matrix:
    # [1 0   0    0     0     ][c0] = [n0]
    # [1 1/4 1/16 1/64  1/256 ][c1]   [n1]
    # [1 1/2 1/4  1/8   1/16  ][c2]   [n2]
    # [1 3/4 9/16 27/64 81/256][c3]   [n3]
    # [1 1   1    1     1     ][c4]   [n4]
    val0 = eval_intersection_polynomial(nodes1, nodes2, 0.0)
    val1 = eval_intersection_polynomial(nodes1, nodes2, 0.25)
    val2 = eval_intersection_polynomial(nodes1, nodes2, 0.5)
    val3 = eval_intersection_polynomial(nodes1, nodes2, 0.75)
    val4 = eval_intersection_polynomial(nodes1, nodes2, 1.0)
    # [c0] =       [ 3   0    0    0    0 ][n0]
    # [c1] = 1 / 3 [-25  48  -36   16  -3 ][n1]
    # [c2] =       [ 70 -208  228 -112  22][n2]
    # [c3] =       [-80  288 -384  224 -48][n3]
    # [c4] =       [ 32 -128  192 -128  32][n4]
    # Since polynomial coefficients, we don't need to divide by 3
    # to get the same polynomial. Avoid the division to avoid round-off.
    return np.asfortranarray(
        [
            3.0 * val0,
            -25.0 * val0
            + 48.0 * val1
            - 36.0 * val2
            + 16.0 * val3
            - 3.0 * val4,
            70.0 * val0
            - 208.0 * val1
            + 228.0 * val2
            - 112.0 * val3
            + 22.0 * val4,
            (
                -80.0 * val0
                + 288.0 * val1
                - 384.0 * val2
                + 224.0 * val3
                - 48.0 * val4
            ),
            32.0 * val0
            - 128.0 * val1
            + 192.0 * val2
            - 128.0 * val3
            + 32.0 * val4,
        ]
    )


def _to_power_basis23(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that the first curve is
    degree two and the second is degree three. In this case, B |eacute| zout's
    `theorem`_ tells us that the **intersection polynomial** is
    degree :math:`2 \cdot 3` hence we return seven coefficients.

    .. note::

       This uses a least-squares fit to the function evaluated at the
       Chebyshev nodes (scaled and shifted onto ``[0, 1]``). Hence, the
       coefficients may be less stable than those produced for smaller
       degrees.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``7``-array of coefficients.
    """
    evaluated = [
        eval_intersection_polynomial(nodes1, nodes2, t_val) for t_val in _CHEB7
    ]
    return polynomial.polyfit(_CHEB7, evaluated, 6)


def _to_power_basis_degree8(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that B |eacute| zout's
    `theorem`_ tells us the **intersection polynomial** is degree
    :math:`8`. This happens if the two curves have degrees one and eight
    or have degrees two and four.

    .. note::

       This uses a least-squares fit to the function evaluated at the
       Chebyshev nodes (scaled and shifted onto ``[0, 1]``). Hence, the
       coefficients may be less stable than those produced for smaller
       degrees.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``9``-array of coefficients.
    """
    evaluated = [
        eval_intersection_polynomial(nodes1, nodes2, t_val) for t_val in _CHEB9
    ]
    return polynomial.polyfit(_CHEB9, evaluated, 8)


def _to_power_basis33(nodes1, nodes2):
    r"""Compute the coefficients of an **intersection polynomial**.

    Helper for :func:`to_power_basis` in the case that each curve is
    degree three. In this case, B |eacute| zout's `theorem`_ tells us
    that the **intersection polynomial** is degree :math:`3 \cdot 3`
    hence we return ten coefficients.

    .. note::

       This uses a least-squares fit to the function evaluated at the
       Chebyshev nodes (scaled and shifted onto ``[0, 1]``). Hence, the
       coefficients may be less stable than those produced for smaller
       degrees.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``10``-array of coefficients.
    """
    evaluated = [
        eval_intersection_polynomial(nodes1, nodes2, t_val)
        for t_val in _CHEB10
    ]
    return polynomial.polyfit(_CHEB10, evaluated, 9)


def to_power_basis(nodes1, nodes2):
    """Compute the coefficients of an **intersection polynomial**.

    .. note::

       This assumes that the degree of the curve given by ``nodes1`` is
       less than or equal to the degree of that given by ``nodes2``.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: Array of coefficients.

    Raises:
        NotImplementedError: If the degree pair is not ``1-1``, ``1-2``,
            ``1-3``, ``1-4``, ``2-2``, ``2-3``, ``2-4`` or ``3-3``.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-return-statements
    _, num_nodes1 = nodes1.shape
    _, num_nodes2 = nodes2.shape
    if num_nodes1 == 2:
        if num_nodes2 == 2:
            return _to_power_basis11(nodes1, nodes2)

        elif num_nodes2 == 3:
            return _to_power_basis12(nodes1, nodes2)

        elif num_nodes2 == 4:
            return _to_power_basis13(nodes1, nodes2)

        elif num_nodes2 == 5:
            return _to_power_basis_degree4(nodes1, nodes2)

    elif num_nodes1 == 3:
        if num_nodes2 == 3:
            return _to_power_basis_degree4(nodes1, nodes2)

        elif num_nodes2 == 4:
            return _to_power_basis23(nodes1, nodes2)

        elif num_nodes2 == 5:
            return _to_power_basis_degree8(nodes1, nodes2)

    elif num_nodes1 == 4:
        if num_nodes2 == 4:
            return _to_power_basis33(nodes1, nodes2)

    raise NotImplementedError(
        "Degree 1",
        num_nodes1 - 1,
        "Degree 2",
        num_nodes2 - 1,
        _POWER_BASIS_ERR,
    )


def polynomial_norm(coeffs):
    r"""Computes :math:`L_2` norm of polynomial on :math:`\left[0, 1\right]`.

    We have

    .. math::

       \left\langle f, f \right\rangle = \sum_{i, j}
           \int_0^1 c_i c_j x^{i + j} \, dx = \sum_{i, j}
           \frac{c_i c_j}{i + j + 1} = \sum_{i} \frac{c_i^2}{2 i + 1}
           + 2 \sum_{j > i} \frac{c_i c_j}{i + j + 1}.

    Args:
        coeffs (numpy.ndarray): ``d + 1``-array of coefficients in monomial /
            power basis.

    Returns:
        float: The :math:`L_2` norm of the polynomial.
    """
    (num_coeffs,) = coeffs.shape
    result = 0.0
    for i in range(num_coeffs):
        coeff_i = coeffs[i]
        result += coeff_i * coeff_i / (2.0 * i + 1.0)
        for j in range(i + 1, num_coeffs):
            coeff_j = coeffs[j]
            result += 2.0 * coeff_i * coeff_j / (i + j + 1.0)
    return np.sqrt(result)


def normalize_polynomial(coeffs, threshold=_L2_THRESHOLD):
    r"""Normalizes a polynomial in the :math:`L_2` sense.

    Does so on the interval :math:\left[0, 1\right]` via
    :func:`polynomial_norm`.

    Args:
        coeffs (numpy.ndarray): ``d + 1``-array of coefficients in monomial /
            power basis.
        threshold (Optional[float]): The point :math:`\tau` below which a
            polynomial will be considered to be numerically equal to zero,
            applies to all :math:`f` with :math`\| f \|_{L_2} < \tau`.

    Returns:
        numpy.ndarray: The normalized polynomial.
    """
    l2_norm = polynomial_norm(coeffs)
    if l2_norm < threshold:
        return np.zeros(coeffs.shape, order="F")

    else:
        coeffs /= l2_norm
        return coeffs


def _get_sigma_coeffs(coeffs):
    r"""Compute "transformed" form of polynomial in Bernstein form.

    Makes sure the "transformed" polynomial is monic (i.e. by dividing by the
    lead coefficient) and just returns the coefficients of the non-lead terms.

    For example, the polynomial

    .. math::

       f(s) = 4 (1 - s)^2 + 3 \cdot 2 s(1 - s) + 7 s^2

    can be transformed into :math:`f(s) = 7 (1 - s)^2 g\left(\sigma\right)`
    for :math:`\sigma = \frac{s}{1 - s}` and :math:`g(\sigma)` a
    polynomial in the power basis:

    .. math::

       g(\sigma) = \frac{4}{7} + \frac{6}{7} \sigma + \sigma^2.

    In cases where terms with "low exponents" of :math:`(1 - s)` have
    coefficient zero, the degree of :math:`g(\sigma)` may not be the
    same as the degree of :math:`f(s)`:

    .. math::

       \begin{align*}
       f(s) &= 5 (1 - s)^4 - 3 \cdot 4 s(1 - s)^3 + 11 \cdot 6 s^2 (1 - s)^2 \\
       \Longrightarrow g(\sigma) &= \frac{5}{66} - \frac{2}{11} \sigma +
           \sigma^2.
       \end{align*}

    Args:
        coeffs (numpy.ndarray): A 1D array of coefficients in
            the Bernstein basis.

    Returns:
        Tuple[Optional[numpy.ndarray], int, int]: A triple of

        * 1D array of the transformed coefficients (will be unset if
          ``effective_degree == 0``)
        * the "full degree" based on the size of ``coeffs``
        * the "effective degree" determined by the number of leading zeros.
    """
    (num_nodes,) = coeffs.shape
    degree = num_nodes - 1
    effective_degree = None
    for index in range(degree, -1, -1):
        if coeffs[index] != 0.0:
            effective_degree = index
            break

    if effective_degree is None:
        # NOTE: This means ``np.all(coeffs == 0.0)``.
        return None, 0, 0

    if effective_degree == 0:
        return None, degree, 0

    sigma_coeffs = coeffs[:effective_degree] / coeffs[effective_degree]
    # Now we need to add the binomial coefficients, but we avoid
    # computing actual binomial coefficients. Starting from the largest
    # exponent, the first ratio of binomial coefficients is
    #       (d C (e - 1)) / (d C e)
    #     = e / (d - e + 1)
    binom_numerator = effective_degree
    binom_denominator = degree - effective_degree + 1
    for exponent in range(effective_degree - 1, -1, -1):
        sigma_coeffs[exponent] *= binom_numerator
        sigma_coeffs[exponent] /= binom_denominator
        # We swap (d C j) with (d C (j - 1)), so `p / q` becomes
        #       (p / q) (d C (j - 1)) / (d C j)
        #     = (p j) / (q (d - j + 1))
        binom_numerator *= exponent
        binom_denominator *= degree - exponent + 1
    return sigma_coeffs, degree, effective_degree


def bernstein_companion(coeffs):
    r"""Compute a companion matrix for a polynomial in Bernstein basis.

    .. note::

       This assumes the caller passes in a 1D array but does not check.

    This takes the polynomial

    .. math::

       f(s) = \sum_{j = 0}^n b_{j, n} \cdot C_j.

    and uses the variable :math:`\sigma = \frac{s}{1 - s}` to rewrite as

    .. math::

       f(s) = (1 - s)^n \sum_{j = 0}^n \binom{n}{j} C_j \sigma^j.

    This converts the Bernstein coefficients :math:`C_j` into "generalized
    Bernstein" coefficients :math:`\widetilde{C_j} = \binom{n}{j} C_j`.

    Args:
        coeffs (numpy.ndarray): A 1D array of coefficients in
            the Bernstein basis.

    Returns:
        Tuple[numpy.ndarray, int, int]: A triple of

        * 2D NumPy array with the companion matrix.
        * the "full degree" based on the size of ``coeffs``
        * the "effective degree" determined by the number of leading zeros.
    """
    sigma_coeffs, degree, effective_degree = _get_sigma_coeffs(coeffs)
    if effective_degree == 0:
        return np.empty((0, 0), order="F"), degree, 0

    companion = np.zeros((effective_degree, effective_degree), order="F")
    # pylint: disable=unsupported-assignment-operation
    companion.flat[
        effective_degree :: effective_degree + 1  # noqa: E203
    ] = 1.0
    # pylint: enable=unsupported-assignment-operation
    companion[0, :] = -sigma_coeffs[::-1]
    return companion, degree, effective_degree


def bezier_roots(coeffs):
    r"""Compute polynomial roots from a polynomial in the Bernstein basis.

    .. note::

       This assumes the caller passes in a 1D array but does not check.

    This takes the polynomial

    .. math::

       f(s) = \sum_{j = 0}^n b_{j, n} \cdot C_j.

    and uses the variable :math:`\sigma = \frac{s}{1 - s}` to rewrite as

    .. math::

       \begin{align*}
       f(s) &= (1 - s)^n \sum_{j = 0}^n \binom{n}{j} C_j \sigma^j \\
            &= (1 - s)^n \sum_{j = 0}^n \widetilde{C_j} \sigma^j.
       \end{align*}

    Then it uses an eigenvalue solver to find the roots of

    .. math::

       g(\sigma) = \sum_{j = 0}^n \widetilde{C_j} \sigma^j

    and convert them back into roots of :math:`f(s)` via
    :math:`s = \frac{\sigma}{1 + \sigma}`.

    For example, consider

    .. math::

       \begin{align*}
       f_0(s) &= 2 (2 - s)(3 + s) \\
              &= 12(1 - s)^2 + 11 \cdot 2s(1 - s) + 8 s^2
       \end{align*}

    First, we compute the companion matrix for

    .. math::

       g_0(\sigma) = 12 + 22 \sigma + 8 \sigma^2

    .. testsetup:: bezier-roots0, bezier-roots1, bezier-roots2

       import numpy as np
       import numpy.linalg
       from bezier._algebraic_intersection import bernstein_companion
       from bezier._algebraic_intersection import bezier_roots

       machine_eps = np.finfo(np.float64).eps

    .. doctest:: bezier-roots0

       >>> coeffs0 = np.asfortranarray([12.0, 11.0, 8.0])
       >>> companion0, _, _ = bernstein_companion(coeffs0)
       >>> companion0
       array([[-2.75, -1.5 ],
              [ 1.  ,  0.  ]])

    then take the eigenvalues of the companion matrix:

    .. doctest:: bezier-roots0

       >>> sigma_values0 = np.linalg.eigvals(companion0)
       >>> sigma_values0
       array([-2.  , -0.75])

    after transforming them, we have the roots of :math:`f(s)`:

    .. doctest:: bezier-roots0

       >>> sigma_values0 / (1.0 + sigma_values0)
       array([ 2., -3.])
       >>> bezier_roots(coeffs0)
       array([ 2., -3.])

    In cases where :math:`s = 1` is a root, the lead coefficient of
    :math:`g` would be :math:`0`, so there is a reduction in the
    companion matrix.

    .. math::

       \begin{align*}
       f_1(s) &= 6 (s - 1)^2 (s - 3) (s - 5) \\
              &= 90 (1 - s)^4 + 33 \cdot 4s(1 - s)^3 + 8 \cdot 6s^2(1 - s)^2
       \end{align*}

    .. doctest:: bezier-roots1
       :options: +NORMALIZE_WHITESPACE

       >>> coeffs1 = np.asfortranarray([90.0, 33.0, 8.0, 0.0, 0.0])
       >>> companion1, degree1, effective_degree1 = bernstein_companion(
       ...     coeffs1)
       >>> companion1
       array([[-2.75 , -1.875],
              [ 1.   ,  0.   ]])
       >>> degree1
       4
       >>> effective_degree1
       2

    so the roots are a combination of the roots determined from
    :math:`s = \frac{\sigma}{1 + \sigma}` and the number of factors
    of :math:`(1 - s)` (i.e. the difference between the degree and
    the effective degree):

    .. doctest:: bezier-roots1

       >>> bezier_roots(coeffs1)
       array([3., 5., 1., 1.])

    In some cases, a polynomial is represented with an "elevated" degree:

    .. math::

       \begin{align*}
       f_2(s) &= 3 (s^2 + 1) \\
              &= 3 (1 - s)^3 + 3 \cdot 3s(1 - s)^2 +
                 4 \cdot 3s^2(1 - s) + 6 s^3
       \end{align*}

    This results in a "point at infinity"
    :math:`\sigma = -1 \Longleftrightarrow s = \infty`:

    .. doctest:: bezier-roots2

       >>> coeffs2 = np.asfortranarray([3.0, 3.0, 4.0, 6.0])
       >>> companion2, _, _ = bernstein_companion(coeffs2)
       >>> companion2
       array([[-2. , -1.5, -0.5],
              [ 1. ,  0. ,  0. ],
              [ 0. ,  1. ,  0. ]])
       >>> sigma_values2 = np.linalg.eigvals(companion2)
       >>> sigma_values2
       array([-1. +0.j , -0.5+0.5j, -0.5-0.5j])


    so we drop any values :math:`\sigma` that are sufficiently close to
    :math:`-1`:

    .. doctest:: bezier-roots2

       >>> expected2 = np.asfortranarray([1.0j, -1.0j])
       >>> roots2 = bezier_roots(coeffs2)
       >>> np.allclose(expected2, roots2, rtol=8 * machine_eps, atol=0.0)
       True

    Args:
        coeffs (numpy.ndarray): A 1D array of coefficients in
            the Bernstein basis.

    Returns:
        numpy.ndarray: A 1D array containing the roots.
    """
    companion, degree, effective_degree = bernstein_companion(coeffs)
    if effective_degree:
        sigma_roots = np.linalg.eigvals(companion)
        # Filter out `sigma = -1`, i.e. "points at infinity".
        # We want the error ||(sigma - (-1))|| ~= 2^{-52}
        to_keep = np.abs(sigma_roots + 1.0) > _SIGMA_THRESHOLD
        sigma_roots = sigma_roots[to_keep]
        s_vals = sigma_roots / (1.0 + sigma_roots)
    else:
        s_vals = np.empty((0,), order="F")
    if effective_degree != degree:
        delta = degree - effective_degree
        s_vals = np.hstack([s_vals, [1] * delta])
    return s_vals


def lu_companion(top_row, value):
    r"""Compute an LU-factored :math:`C - t I` and its 1-norm.

    .. _dgecon:
        http://www.netlib.org/lapack/explore-html/dd/d9a/group__double_g_ecomputational_ga188b8d30443d14b1a3f7f8331d87ae60.html#ga188b8d30443d14b1a3f7f8331d87ae60
    .. _dgetrf:
        http://www.netlib.org/lapack/explore-html/dd/d9a/group__double_g_ecomputational_ga0019443faea08275ca60a734d0593e60.html#ga0019443faea08275ca60a734d0593e60

    .. note::

       The output of this function is intended to be used with `dgecon`_
       from LAPACK. ``dgecon`` expects both the 1-norm of the matrix
       and expects the matrix to be passed in an already LU-factored form
       (via `dgetrf`_).

    The companion matrix :math:`C` is given by the ``top_row``, for
    example, the polynomial :math:`t^3 + 3 t^2 - t + 2` has
    a top row of ``-3, 1, -2`` and the corresponding companion matrix
    is:

    .. math::

       \left[\begin{array}{c c c}
         -3 & 1 & -2 \\
          1 & 0 &  0 \\
          0 & 1 &  0
       \end{array}\right]

    After doing a full cycle of the rows (shifting the first to the last and
    moving all other rows up), row reduction of :math:`C - t I` yields

    .. math::

       \left[\begin{array}{c c c}
          1     & -t &  0 \\
          0     &  1 & -t \\
         -3 - t &  1 & -2
       \end{array}\right] =
       \left[\begin{array}{c c c}
          1     & 0             & 0 \\
          0     & 1             & 0 \\
         -3 - t & 1 + t(-3 - t) & 1
       \end{array}\right]
       \left[\begin{array}{c c c}
          1 & -t &  0                    \\
          0 &  1 & -t                    \\
          0 &  0 & -2 + t(1 + t(-3 - t))
       \end{array}\right]

    and in general, the terms in the bottom row correspond to the intermediate
    values involved in evaluating the polynomial via `Horner's method`_.

    .. _Horner's method: https://en.wikipedia.org/wiki/Horner%27s_method

    .. testsetup:: lu-companion

       import numpy as np
       import numpy.linalg
       from bezier._algebraic_intersection import lu_companion

    .. doctest:: lu-companion

       >>> top_row = np.asfortranarray([-3.0, 1.0, -2.0])
       >>> t_val = 0.5
       >>> lu_mat, one_norm = lu_companion(top_row, t_val)
       >>> lu_mat
       array([[ 1.   , -0.5  ,  0.   ],
              [ 0.   ,  1.   , -0.5  ],
              [-3.5  , -0.75 , -2.375]])
       >>> one_norm
       4.5
       >>> l_mat = np.tril(lu_mat, k=-1) + np.eye(3)
       >>> u_mat = np.triu(lu_mat)
       >>> a_mat = l_mat.dot(u_mat)
       >>> a_mat
       array([[ 1. , -0.5,  0. ],
              [ 0. ,  1. , -0.5],
              [-3.5,  1. , -2. ]])
       >>> np.linalg.norm(a_mat, ord=1)
       4.5

    Args:
        top_row (numpy.ndarray): 1D array, top row of companion matrix.
        value (float): The :math:`t` value used to form :math:`C - t I`.

    Returns:
        Tuple[numpy.ndarray, float]: Pair of

        * 2D array of LU-factored form of :math:`C - t I`, with the
          non-diagonal part of :math:`L` stored in the strictly lower triangle
          and :math:`U` stored in the upper triangle (we skip the permutation
          matrix, as it won't impact the 1-norm)
        * the 1-norm the matrix :math:`C - t I`

        As mentioned above, these two values are meant to be used with
        `dgecon`_.
    """
    (degree,) = top_row.shape
    lu_mat = np.zeros((degree, degree), order="F")
    if degree == 1:
        lu_mat[0, 0] = top_row[0] - value
        return lu_mat, abs(lu_mat[0, 0])

    # Column 0: Special case since it doesn't have ``-t`` above the diagonal.
    horner_curr = top_row[0] - value
    one_norm = 1.0 + abs(horner_curr)
    lu_mat[0, 0] = 1.0
    lu_mat[degree - 1, 0] = horner_curr
    # Columns 1-(end - 1): Three values in LU and C - t I.
    abs_one_plus = 1.0 + abs(value)
    last_row = degree - 1
    for col in range(1, degree - 1):
        curr_coeff = top_row[col]
        horner_curr = value * horner_curr + curr_coeff
        one_norm = max(one_norm, abs_one_plus + abs(curr_coeff))
        lu_mat[col - 1, col] = -value
        lu_mat[col, col] = 1.0
        lu_mat[last_row, col] = horner_curr
    # Last Column: Special case since it doesn't have ``-1`` on the diagonal.
    curr_coeff = top_row[last_row]
    horner_curr = value * horner_curr + curr_coeff
    one_norm = max(one_norm, abs(value) + abs(curr_coeff))
    lu_mat[last_row - 1, last_row] = -value
    lu_mat[last_row, last_row] = horner_curr
    return lu_mat, one_norm


def _reciprocal_condition_number(lu_mat, one_norm):
    r"""Compute reciprocal condition number of a matrix.

    Args:
        lu_mat (numpy.ndarray): A 2D array of a matrix :math:`A` that has been
            LU-factored, with the non-diagonal part of :math:`L` stored in the
            strictly lower triangle and :math:`U` stored in the upper triangle.
        one_norm (float): The 1-norm of the original matrix :math:`A`.

    Returns:
        float: The reciprocal condition number of :math:`A`.

    Raises:
        OSError: If SciPy is not installed.
        RuntimeError: If the reciprocal 1-norm condition number could not
            be computed.
    """
    if _scipy_lapack is None:
        raise OSError("This function requires SciPy for calling into LAPACK.")

    # pylint: disable=no-member
    rcond, info = _scipy_lapack.dgecon(lu_mat, one_norm)
    # pylint: enable=no-member
    if info != 0:
        raise RuntimeError(
            "The reciprocal 1-norm condition number could not be computed."
        )

    return rcond


def bezier_value_check(coeffs, s_val, rhs_val=0.0):
    r"""Check if a polynomial in the Bernstein basis evaluates to a value.

    This is intended to be used for root checking, i.e. for a polynomial
    :math:`f(s)` and a particular value :math:`s_{\ast}`:

       Is it true that :math:`f\left(s_{\ast}\right) = 0`?

    Does so by re-stating as a matrix rank problem. As in
    :func:`~bezier._algebraic_intersection.bezier_roots`, we can rewrite

    .. math::

       f(s) = (1 - s)^n g\left(\sigma\right)

    for :math:`\sigma = \frac{s}{1 - s}` and :math:`g\left(\sigma\right)`
    written in the power / monomial basis. Now, checking if
    :math:`g\left(\sigma\right) = 0` is a matter of checking that
    :math:`\det\left(C_g - \sigma I\right) = 0` (where :math:`C_g` is the
    companion matrix of :math:`g`).

    Due to issues of numerical stability, we'd rather ask if
    :math:`C_g - \sigma I` is singular to numerical precision. A typical
    approach to this using the singular values (assuming :math:`C_g` is
    :math:`m \times m`) is that the matrix is singular if

    .. math::

       \sigma_m < m \varepsilon \sigma_1 \Longleftrightarrow
           \frac{1}{\kappa_2} < m \varepsilon

    (where :math:`\kappa_2` is the 2-norm condition number of the matrix).
    Since we also know that :math:`\kappa_2 < \kappa_1`, a stronger
    requirement would be

    .. math::

       \frac{1}{\kappa_1} < \frac{1}{\kappa_2} < m \varepsilon.

    This is more useful since it is **much** easier to compute the 1-norm
    condition number / reciprocal condition number (and methods in LAPACK are
    provided for doing so).

    Args:
        coeffs (numpy.ndarray): A 1D array of coefficients in
            the Bernstein basis representing a polynomial.
        s_val (float): The value to check on the polynomial:
            :math:`f(s) = r`.
        rhs_val (Optional[float]): The value to check that the polynomial
            evaluates to. Defaults to ``0.0``.

    Returns:
        bool: Indicates if :math:`f\left(s_{\ast}\right) = r` (where
        :math:`s_{\ast}` is ``s_val`` and :math:`r` is ``rhs_val``).
    """
    if s_val == 1.0:
        return coeffs[-1] == rhs_val

    shifted_coeffs = coeffs - rhs_val
    sigma_coeffs, _, effective_degree = _get_sigma_coeffs(shifted_coeffs)
    if effective_degree == 0:
        # This means that all coefficients except the ``(1 - s)^n``
        # term are zero, so we have ``f(s) = C (1 - s)^n``. Since we know
        # ``s != 1``, this can only be zero if ``C == 0``.
        return shifted_coeffs[0] == 0.0

    sigma_val = s_val / (1.0 - s_val)
    lu_mat, one_norm = lu_companion(-sigma_coeffs[::-1], sigma_val)
    rcond = _reciprocal_condition_number(lu_mat, one_norm)
    # "Is a root?" IFF Singular IF ``1/kappa_1 < m epsilon``
    return rcond < effective_degree * _SINGULAR_EPS


def roots_in_unit_interval(coeffs):
    r"""Compute roots of a polynomial in the unit interval.

    Args:
        coeffs (numpy.ndarray): A 1D array (size ``d + 1``) of coefficients in
            monomial / power basis.

    Returns:
        numpy.ndarray: ``N``-array of real values in :math:`\left[0, 1\right]`.
    """
    all_roots = polynomial.polyroots(coeffs)
    # Only keep roots inside or very near to the unit interval.
    all_roots = all_roots[
        (_UNIT_INTERVAL_WIGGLE_START < all_roots.real)
        & (all_roots.real < _UNIT_INTERVAL_WIGGLE_END)
    ]
    # Only keep roots with very small imaginary part. (Really only
    # keep the real parts.)
    real_inds = np.abs(all_roots.imag) < _IMAGINARY_WIGGLE
    return all_roots[real_inds].real


def _strip_leading_zeros(coeffs, threshold=_COEFFICIENT_THRESHOLD):
    r"""Strip leading zero coefficients from a polynomial.

    .. note::

       This assumes the polynomial :math:`f` defined by ``coeffs``
       has been normalized (via :func:`.normalize_polynomial`).

    Args:
        coeffs (numpy.ndarray): ``d + 1``-array of coefficients in monomial /
            power basis.
        threshold (Optional[float]): The point :math:`\tau` below which a
            a coefficient will be considered to be numerically zero.

    Returns:
        numpy.ndarray: The same coefficients without any unnecessary zero
        terms.
    """
    while np.abs(coeffs[-1]) < threshold:
        coeffs = coeffs[:-1]
    return coeffs


def _check_non_simple(coeffs):
    r"""Checks that a polynomial has no non-simple roots.

    Does so by computing the companion matrix :math:`A` of :math:`f'`
    and then evaluating the rank of :math:`B = f(A)`. If :math:`B` is not
    full rank, then :math:`f` and :math:`f'` have a shared factor.

    See: https://dx.doi.org/10.1016/0024-3795(70)90023-6

    .. note::

       This assumes that :math:`f \neq 0`.

    Args:
        coeffs (numpy.ndarray): ``d + 1``-array of coefficients in monomial /
            power basis.

    Raises:
        NotImplementedError: If the polynomial has non-simple roots.
    """
    coeffs = _strip_leading_zeros(coeffs)
    (num_coeffs,) = coeffs.shape
    if num_coeffs < 3:
        return

    deriv_poly = polynomial.polyder(coeffs)
    companion = polynomial.polycompanion(deriv_poly)
    # NOTE: ``polycompanion()`` returns a C-contiguous array.
    companion = companion.T
    # Use Horner's method to evaluate f(companion)
    num_companion, _ = companion.shape
    id_mat = np.eye(num_companion, order="F")
    evaluated = coeffs[-1] * id_mat
    for index in range(num_coeffs - 2, -1, -1):
        coeff = coeffs[index]
        evaluated = (
            _py_helpers.matrix_product(evaluated, companion) + coeff * id_mat
        )
    if num_companion == 1:
        # NOTE: This relies on the fact that coeffs is normalized.
        if np.abs(evaluated[0, 0]) > _NON_SIMPLE_THRESHOLD:
            rank = 1
        else:
            rank = 0
    else:
        rank = np.linalg.matrix_rank(evaluated)
    if rank < num_companion:
        raise NotImplementedError(_NON_SIMPLE_ERR, coeffs)


def _resolve_and_add(nodes1, s_val, final_s, nodes2, t_val, final_t):
    """Resolve a computed intersection and add to lists.

    We perform one Newton step to deal with any residual issues of
    high-degree polynomial solves (one of which depends on the already
    approximate ``x_val, y_val``).

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        s_val (float): The approximate intersection parameter
            along ``nodes1``.
        final_s (List[float]): The list of accepted intersection
            parameters ``s``.
        nodes2 (numpy.ndarray): The nodes in the second curve.
        t_val (float): The approximate intersection parameter
            along ``nodes2``.
        final_t (List[float]): The list of accepted intersection
            parameters ``t``.
    """
    s_val, t_val = _intersection_helpers.newton_refine(
        s_val, nodes1, t_val, nodes2
    )
    s_val, success_s = _helpers.wiggle_interval(s_val)
    t_val, success_t = _helpers.wiggle_interval(t_val)
    if not (success_s and success_t):
        return

    final_s.append(s_val)
    final_t.append(t_val)


def intersect_curves(nodes1, nodes2):
    r"""Intersect two parametric B |eacute| zier curves.

    Args:
        nodes1 (numpy.ndarray): The nodes in the first curve.
        nodes2 (numpy.ndarray): The nodes in the second curve.

    Returns:
        numpy.ndarray: ``2 x N`` array of intersection parameters.
        Each row contains a pair of values :math:`s` and :math:`t`
        (each in :math:`\left[0, 1\right]`) such that the curves
        intersect: :math:`B_1(s) = B_2(t)`.

    Raises:
        NotImplementedError: If the "intersection polynomial" is
        all zeros -- which indicates coincident curves.
    """
    nodes1 = _curve_helpers.full_reduce(nodes1)
    nodes2 = _curve_helpers.full_reduce(nodes2)
    _, num_nodes1 = nodes1.shape
    _, num_nodes2 = nodes2.shape
    swapped = False
    if num_nodes1 > num_nodes2:
        nodes1, nodes2 = nodes2, nodes1
        swapped = True
    coeffs = normalize_polynomial(to_power_basis(nodes1, nodes2))
    if np.all(coeffs == 0.0):
        raise NotImplementedError(_COINCIDENT_ERR)

    _check_non_simple(coeffs)
    t_vals = roots_in_unit_interval(coeffs)
    final_s = []
    final_t = []
    for t_val in t_vals:
        (x_val,), (y_val,) = _curve_helpers.evaluate_multi(
            nodes2, np.asfortranarray([t_val])
        )
        s_val = locate_point(nodes1, x_val, y_val)
        if s_val is not None:
            _resolve_and_add(nodes1, s_val, final_s, nodes2, t_val, final_t)
    result = np.zeros((2, len(final_s)), order="F")
    if swapped:
        final_s, final_t = final_t, final_s
    result[0, :] = final_s
    result[1, :] = final_t
    return result


def poly_to_power_basis(bezier_coeffs):
    """Convert a B |eacute| zier curve to polynomial in power basis.

    .. note::

       This assumes, but does not verify, that the "B |eacute| zier
       degree" matches the true degree of the curve. Callers can
       guarantee this by calling :func:`.full_reduce`.

    Args:
        bezier_coeffs (numpy.ndarray): A 1D array of coefficients in
            the Bernstein basis.

    Returns:
        numpy.ndarray: 1D array of coefficients in monomial basis.

    Raises:
        .UnsupportedDegree: If the degree of the curve is not among
            0, 1, 2 or 3.
    """
    (num_coeffs,) = bezier_coeffs.shape
    if num_coeffs == 1:
        return bezier_coeffs

    elif num_coeffs == 2:
        # C0 (1 - s) + C1 s = C0 + (C1 - C0) s
        coeff0, coeff1 = bezier_coeffs
        return np.asfortranarray([coeff0, coeff1 - coeff0])

    elif num_coeffs == 3:
        #   C0 (1 - s)^2 + C1 2 (1 - s) s + C2 s^2
        # = C0 + 2(C1 - C0) s + (C2 - 2 C1 + C0) s^2
        coeff0, coeff1, coeff2 = bezier_coeffs
        return np.asfortranarray(
            [coeff0, 2.0 * (coeff1 - coeff0), coeff2 - 2.0 * coeff1 + coeff0]
        )

    elif num_coeffs == 4:
        #   C0 (1 - s)^3 + C1 3 (1 - s)^2 + C2 3 (1 - s) s^2 + C3 s^3
        # = C0 + 3(C1 - C0) s + 3(C2 - 2 C1 + C0) s^2 +
        #   (C3 - 3 C2 + 3 C1 - C0) s^3
        coeff0, coeff1, coeff2, coeff3 = bezier_coeffs
        return np.asfortranarray(
            [
                coeff0,
                3.0 * (coeff1 - coeff0),
                3.0 * (coeff2 - 2.0 * coeff1 + coeff0),
                coeff3 - 3.0 * coeff2 + 3.0 * coeff1 - coeff0,
            ]
        )

    else:
        raise _py_helpers.UnsupportedDegree(
            num_coeffs - 1, supported=(0, 1, 2, 3)
        )


def locate_point(nodes, x_val, y_val):
    r"""Find the parameter corresponding to a point on a curve.

    .. note::

       This assumes that the curve :math:`B(s, t)` defined by ``nodes``
       lives in :math:`\mathbf{R}^2`.

    Args:
        nodes (numpy.ndarray): The nodes defining a B |eacute| zier curve.
        x_val (float): The :math:`x`-coordinate of the point.
        y_val (float): The :math:`y`-coordinate of the point.

    Returns:
        Optional[float]: The parameter on the curve (if it exists).
    """
    # First, reduce to the true degree of x(s) and y(s).
    zero1 = _curve_helpers.full_reduce(nodes[[0], :]) - x_val
    zero2 = _curve_helpers.full_reduce(nodes[[1], :]) - y_val
    # Make sure we have the lowest degree in front, to make the polynomial
    # solve have the fewest number of roots.
    if zero1.shape[1] > zero2.shape[1]:
        zero1, zero2 = zero2, zero1
    # If the "smallest" is a constant, we can't find any roots from it.
    if zero1.shape[1] == 1:
        # NOTE: We assume that callers won't pass ``nodes`` that are
        #       degree 0, so if ``zero1`` is a constant, ``zero2`` won't be.
        zero1, zero2 = zero2, zero1
    power_basis1 = poly_to_power_basis(zero1[0, :])
    all_roots = roots_in_unit_interval(power_basis1)
    if all_roots.size == 0:
        return None

    # NOTE: We normalize ``power_basis2`` because we want to check for
    #       "zero" values, i.e. f2(s) == 0.
    power_basis2 = normalize_polynomial(poly_to_power_basis(zero2[0, :]))
    near_zero = np.abs(polynomial.polyval(all_roots, power_basis2))
    index = np.argmin(near_zero)
    if near_zero[index] < _ZERO_THRESHOLD:
        return all_roots[index]

    return None


def all_intersections(nodes_first, nodes_second):
    r"""Find the points of intersection among a pair of curves.

    .. note::

       This assumes both curves are :math:`\mathbf{R}^2`, but does not
       **explicitly** check this. However, functions used here will fail if
       that assumption fails.

    Args:
        nodes_first (numpy.ndarray): Control points of a curve to be
            intersected with ``nodes_second``.
        nodes_second (numpy.ndarray): Control points of a curve to be
            intersected with ``nodes_first``.

    Returns:
        Tuple[numpy.ndarray, bool]: An array and a flag:

        * A ``2 x N`` array of intersection parameters.
          Each row contains a pair of values :math:`s` and :math:`t`
          (each in :math:`\left[0, 1\right]`) such that the curves
          intersect: :math:`B_1(s) = B_2(t)`.
        * Flag indicating if the curves are coincident. (For now, this
          will always be :data:`False` since :func:`.intersect_curves`
          fails if coincident curves are detected.)
    """
    # Only attempt this if the bounding boxes intersect.
    bbox_int = _geometric_intersection.bbox_intersect(
        nodes_first, nodes_second
    )
    if bbox_int == _DISJOINT:
        return np.empty((2, 0), order="F"), False

    return intersect_curves(nodes_first, nodes_second), False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for B |eacute| zier Triangles / Triangles.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

.. testsetup:: *

   import numpy as np
   import bezier
"""

import numpy as np

from bezier import _base
from bezier import _curve_helpers
from bezier import _plot_helpers
from bezier import _py_helpers
from bezier import _py_intersection_helpers
from bezier import _py_triangle_helpers
from bezier import _py_triangle_intersection
from bezier import _symbolic
from bezier import _triangle_helpers
from bezier import _triangle_intersection
from bezier import curve as _curve_mod
from bezier import curved_polygon


_SIGN = np.sign  # pylint: disable=no-member
_LOCATE_ERROR_TEMPLATE = (
    "Dimension mismatch: This triangle is {:d}-dimensional, "
    "so the point should be a {:d} x 1 NumPy array. "
    "Instead the point {} has dimensions {}."
)
_STRATEGY = _py_intersection_helpers.IntersectionStrategy


class Triangle(_base.Base):
    r"""Represents a B |eacute| zier `triangle`_.

    .. _triangle: https://en.wikipedia.org/wiki/B%C3%A9zier_triangle
    .. _unit simplex:
        https://en.wikipedia.org/wiki/Simplex#The_standard_simplex
    .. _barycentric coordinates:
        https://en.wikipedia.org/wiki/Barycentric_coordinate_system

    We define a B |eacute| zier triangle as a mapping from the
    `unit simplex`_ in :math:`\mathbf{R}^2` (i.e. the unit triangle) onto a
    triangle in an arbitrary dimension. We use `barycentric coordinates`_

    .. math::

       \lambda_1 = 1 - s - t, \lambda_2 = s, \lambda_3 = t

    for points in the unit triangle
    :math:`\left\{(s, t) \mid 0 \leq s, t, s + t \leq 1\right\}`:

    .. image:: ../../images/unit_triangle.png
       :align: center

    As with curves, using these weights we get convex combinations
    of points :math:`v_{i, j, k}` in some vector space:

    .. math::

       B\left(\lambda_1, \lambda_2, \lambda_3\right) =
           \sum_{i + j + k = d} \binom{d}{i \, j \, k}
           \lambda_1^i \lambda_2^j \lambda_3^k \cdot v_{i, j, k}

    .. note::

       We assume the nodes are ordered from left-to-right and from
       bottom-to-top. So for example, the linear triangle:

       .. code-block:: rest

          (0,0,1)

          (1,0,0)  (0,1,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c}
              v_{1,0,0} & v_{0,1,0} & v_{0,0,1} \end{array}\right]

       the quadratic triangle:

       .. code-block:: rest

          (0,0,2)

          (1,0,1)  (0,1,1)

          (2,0,0)  (1,1,0)  (0,2,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c c c c}
              v_{2,0,0} & v_{1,1,0} &
              v_{0,2,0} & v_{1,0,1} &
              v_{0,1,1} & v_{0,0,2} \end{array}\right]

       the cubic triangle:

       .. code-block:: rest

          (0,0,3)

          (1,0,2)  (0,1,2)

          (2,0,1)  (1,1,1)  (0,2,1)

          (3,0,0)  (2,1,0)  (1,2,0)  (0,3,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c c c c c c c c}
              v_{3,0,0} & v_{2,1,0} &
              v_{1,2,0} & v_{0,3,0} &
              v_{2,0,1} & v_{1,1,1} &
              v_{0,2,1} & v_{1,0,2} &
              v_{0,1,2} & v_{0,0,3} \end{array}\right]

       and so on.

       The index formula

       .. math::

          j + \frac{k}{2} \left(2 (i + j) + k + 3\right)

       can be used to map a triple :math:`(i, j, k)` onto the
       corresponding linear index, but it is not particularly insightful
       or useful.

    .. image:: ../../images/triangle_constructor.png
       :align: center

    .. doctest:: triangle-constructor

       >>> import bezier
       >>> nodes = np.asfortranarray([
       ...     [0.0, 0.5, 1.0 , 0.125, 0.375, 0.25],
       ...     [0.0, 0.0, 0.25, 0.5  , 0.375, 1.0 ],
       ... ])
       >>> triangle = bezier.Triangle(nodes, degree=2)
       >>> triangle
       <Triangle (degree=2, dimension=2)>

    .. testcleanup:: triangle-constructor

       import make_images
       make_images.unit_triangle()
       make_images.triangle_constructor(triangle)

    Args:
        nodes (Sequence[Sequence[numbers.Number]]): The nodes in the triangle.
            Must be convertible to a 2D NumPy array of floating point values,
            where the columns represent each node while the rows are the
            dimension of the ambient space.
        degree (int): The degree of the triangle. This is assumed to
            correctly correspond to the number of ``nodes``. Use
            :meth:`from_nodes` if the degree has not yet been computed.
        copy (bool): Flag indicating if the nodes should be copied before
            being stored. Defaults to :data:`True` since callers may
            freely mutate ``nodes`` after passing in.
        verify (bool): Flag indicating if the degree should be verified against
            the number of nodes. Defaults to :data:`True`.
    """

    __slots__ = (
        "_dimension",  # From base class
        "_nodes",  # From base class
        "_degree",  # From constructor
        "_edges",  # Empty default
    )

    def __init__(self, nodes, degree, *, copy=True, verify=True):
        super(Triangle, self).__init__(nodes, copy=copy)
        self._degree = degree
        self._edges = None
        self._verify_degree(verify)

    @classmethod
    def from_nodes(cls, nodes, copy=True):
        """Create a :class:`.Triangle` from nodes.

        Computes the ``degree`` based on the shape of ``nodes``.

        Args:
            nodes (Sequence[Sequence[numbers.Number]]): The nodes in the
                triangle. Must be convertible to a 2D NumPy array of floating
                point values, where the columns represent each node while the
                rows are the dimension of the ambient space.
            copy (bool): Flag indicating if the nodes should be copied before
                being stored. Defaults to :data:`True` since callers may
                freely mutate ``nodes`` after passing in.

        Returns:
            Triangle: The constructed triangle.
        """
        nodes_np = _base.sequence_to_array(nodes)
        _, num_nodes = nodes_np.shape
        degree = cls._get_degree(num_nodes)
        # NOTE: **Explicitly** verify because ``_get_degree`` does not.
        return cls(nodes_np, degree, copy=copy, verify=True)

    @staticmethod
    def _get_degree(num_nodes):
        """Get the degree of the current triangle.

        .. note::

            If ``num_nodes`` isn't a triangular number, no degree can be
            correct so the return value will be invalid. Callers should use
            ``verify`` in the constructor to ensure correctness.

        Args:
            num_nodes (int): The number of control points for a
                B |eacute| zier triangle.

        Returns:
            int: The degree :math:`d` such that :math:`(d + 1)(d + 2)/2`
            equals ``num_nodes``.
        """
        # 8 * num_nodes = 4(d + 1)(d + 2)
        #               = 4d^2 + 12d + 8
        #               = (2d + 3)^2 - 1
        d_float = 0.5 * (np.sqrt(8.0 * num_nodes + 1.0) - 3.0)
        return int(np.round(d_float))

    def _verify_degree(self, verify):
        """Verify that the number of nodes matches the degree.

        Args:
            verify (bool): Flag indicating if the degree should be verified
                against the number of nodes.

        Raises:
            ValueError: If ``verify`` is :data:`True` and the number of nodes
                does not match the degree.
        """
        if not verify:
            return

        _, num_nodes = self._nodes.shape
        twice_expected_nodes = (self._degree + 1) * (self._degree + 2)
        # Avoid rounding by division by 2.
        if twice_expected_nodes == 2 * num_nodes:
            return

        msg = (
            f"A degree {self._degree} triangle should have "
            f"{0.5 * twice_expected_nodes:g} nodes, not {num_nodes}."
        )
        raise ValueError(msg)

    @property
    def area(self):
        r"""The area of the current triangle.

        For triangles in :math:`\mathbf{R}^2`, this computes the area via
        Green's theorem. Using the vector field :math:`\mathbf{F} =
        \left[-y, x\right]^T`, since :math:`\partial_x(x) - \partial_y(-y) = 2`
        Green's theorem says twice the area is equal to

        .. math::

           \int_{B\left(\mathcal{U}\right)} 2 \, d\mathbf{x} =
           \int_{\partial B\left(\mathcal{U}\right)} -y \, dx + x \, dy.

        This relies on the assumption that the current triangle is valid, which
        implies that the image of the unit triangle under the B |eacute| zier
        map --- :math:`B\left(\mathcal{U}\right)`  --- has the edges of the
        triangle as its boundary.

        Note that for a given edge :math:`C(r)` with control points
        :math:`x_j, y_j`, the integral can be simplified:

        .. math::

            \int_C -y \, dx + x \, dy = \int_0^1 (x y' - y x') \, dr
                = \sum_{i < j} (x_i y_j - y_i x_j) \int_0^1 b_{i, d}
                b'_{j, d} \, dr

        where :math:`b_{i, d}, b_{j, d}` are Bernstein basis polynomials.

        Returns:
            float: The area of the current triangle.

        Raises:
            NotImplementedError: If the current triangle isn't in
                :math:`\mathbf{R}^2`.
        """
        if self._dimension != 2:
            raise NotImplementedError(
                "2D is the only supported dimension",
                "Current dimension",
                self._dimension,
            )

        edge1, edge2, edge3 = self._get_edges()
        return _triangle_helpers.compute_area(
            (edge1._nodes, edge2._nodes, edge3._nodes)
        )

    def _compute_edges(self):
        """Compute the edges of the current triangle.

        Returns:
            Tuple[~curve.Curve, ~curve.Curve, ~curve.Curve]: The edges of
            the triangle.
        """
        nodes1, nodes2, nodes3 = _triangle_helpers.compute_edge_nodes(
            self._nodes, self._degree
        )
        edge1 = _curve_mod.Curve(
            nodes1, self._degree, copy=False, verify=False
        )
        edge2 = _curve_mod.Curve(
            nodes2, self._degree, copy=False, verify=False
        )
        edge3 = _curve_mod.Curve(
            nodes3, self._degree, copy=False, verify=False
        )
        return edge1, edge2, edge3

    def _get_edges(self):
        """Get the edges for the current triangle.

        If they haven't been computed yet, first compute and store them.

        This is provided as a means for internal calls to get the edges
        without copying (since :attr:`.edges` copies before giving to
        a user to keep the stored data immutable).

        Returns:
            Tuple[~bezier.curve.Curve, ~bezier.curve.Curve, \
                  ~bezier.curve.Curve]: The edges of
            the triangle.
        """
        if self._edges is None:
            self._edges = self._compute_edges()
        return self._edges

    @property
    def edges(self):
        """The edges of the triangle.

        .. doctest:: triangle-edges
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0,  0.5   , 1.0, 0.1875, 0.625, 0.0],
           ...     [0.0, -0.1875, 0.0, 0.5   , 0.625, 1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> edge1, _, _ = triangle.edges
           >>> edge1
           <Curve (degree=2, dimension=2)>
           >>> edge1.nodes
           array([[ 0. ,  0.5   , 1. ],
                  [ 0. , -0.1875, 0. ]])

        Returns:
            Tuple[~bezier.curve.Curve, ~bezier.curve.Curve, \
                  ~bezier.curve.Curve]: The edges of
            the triangle.
        """
        edge1, edge2, edge3 = self._get_edges()
        # NOTE: It is crucial that we return copies here. Since the edges
        #       are cached, if they were mutable, callers could
        #       inadvertently mutate the cached value.
        edge1 = edge1.copy()
        edge2 = edge2.copy()
        edge3 = edge3.copy()
        return edge1, edge2, edge3

    @staticmethod
    def _verify_barycentric(lambda1, lambda2, lambda3):
        """Verifies that weights are barycentric and on the reference triangle.

        I.e., checks that they sum to one and are all non-negative.

        Args:
            lambda1 (float): Parameter along the reference triangle.
            lambda2 (float): Parameter along the reference triangle.
            lambda3 (float): Parameter along the reference triangle.

        Raises:
            ValueError: If the weights are not valid barycentric
                coordinates, i.e. they don't sum to ``1``.
            ValueError: If some weights are negative.
        """
        weights_total = lambda1 + lambda2 + lambda3
        if not np.allclose(weights_total, 1.0, atol=0.0):
            raise ValueError(
                "Weights do not sum to 1", lambda1, lambda2, lambda3
            )

        if lambda1 < 0.0 or lambda2 < 0.0 or lambda3 < 0.0:
            raise ValueError(
                "Weights must be positive", lambda1, lambda2, lambda3
            )

    def evaluate_barycentric(self, lambda1, lambda2, lambda3, _verify=True):
        r"""Compute a point on the triangle.

        Evaluates :math:`B\left(\lambda_1, \lambda_2, \lambda_3\right)`.

        .. image:: ../../images/triangle_evaluate_barycentric.png
           :align: center

        .. testsetup:: triangle-barycentric, triangle-barycentric-fail1,
                       triangle-barycentric-fail2,
                       triangle-barycentric-no-verify

           import numpy as np
           import bezier
           nodes = np.asfortranarray([
               [0.0, 0.5, 1.0 , 0.125, 0.375, 0.25],
               [0.0, 0.0, 0.25, 0.5  , 0.375, 1.0 ],
           ])
           triangle = bezier.Triangle(nodes, degree=2)

        .. doctest:: triangle-barycentric
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5, 1.0 , 0.125, 0.375, 0.25],
           ...     [0.0, 0.0, 0.25, 0.5  , 0.375, 1.0 ],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> point = triangle.evaluate_barycentric(0.125, 0.125, 0.75)
           >>> point
           array([[0.265625  ],
                  [0.73046875]])

        .. testcleanup:: triangle-barycentric

           import make_images
           make_images.triangle_evaluate_barycentric(triangle, point)

        However, this can't be used for points **outside** the
        reference triangle:

        .. doctest:: triangle-barycentric-fail1

           >>> triangle.evaluate_barycentric(-0.25, 0.75, 0.5)
           Traceback (most recent call last):
             ...
           ValueError: ('Weights must be positive', -0.25, 0.75, 0.5)

        or for non-barycentric coordinates;

        .. doctest:: triangle-barycentric-fail2

           >>> triangle.evaluate_barycentric(0.25, 0.25, 0.25)
           Traceback (most recent call last):
             ...
           ValueError: ('Weights do not sum to 1', 0.25, 0.25, 0.25)

        However, these "invalid" inputs can be used if ``_verify`` is
        :data:`False`.

        .. doctest:: triangle-barycentric-no-verify
           :options: +NORMALIZE_WHITESPACE

           >>> triangle.evaluate_barycentric(-0.25, 0.75, 0.5, _verify=False)
           array([[0.6875  ],
                  [0.546875]])
           >>> triangle.evaluate_barycentric(0.25, 0.25, 0.25, _verify=False)
           array([[0.203125],
                  [0.1875  ]])

        Args:
            lambda1 (float): Parameter along the reference triangle.
            lambda2 (float): Parameter along the reference triangle.
            lambda3 (float): Parameter along the reference triangle.
            _verify (Optional[bool]): Indicates if the barycentric coordinates
                should be verified as summing to one and all non-negative (i.e.
                verified as barycentric). Can either be used to evaluate at
                points outside the domain, or to save time when the caller
                already knows the input is verified. Defaults to :data:`True`.

        Returns:
            numpy.ndarray: The point on the triangle (as a two dimensional
            NumPy array with a single column).

        Raises:
            ValueError: If the weights are not valid barycentric
                coordinates, i.e. they don't sum to ``1``. (Won't raise if
                ``_verify=False``.)
            ValueError: If some weights are negative. (Won't raise if
                ``_verify=False``.)
        """
        if _verify:
            self._verify_barycentric(lambda1, lambda2, lambda3)
        return _triangle_helpers.evaluate_barycentric(
            self._nodes, self._degree, lambda1, lambda2, lambda3
        )

    def evaluate_barycentric_multi(self, param_vals, _verify=True):
        r"""Compute multiple points on the triangle.

        Assumes ``param_vals`` has three columns of barycentric coordinates.
        See :meth:`evaluate_barycentric` for more details on how each row of
        parameter values is evaluated.

        .. image:: ../../images/triangle_evaluate_barycentric_multi.png
           :align: center

        .. doctest:: triangle-eval-multi2
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.0 , 2.0, -1.5, -0.5, -3.0],
           ...     [0.0, 0.75, 1.0,  1.0,  1.5,  2.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> triangle
           <Triangle (degree=2, dimension=2)>
           >>> param_vals = np.asfortranarray([
           ...     [0.   , 0.25, 0.75 ],
           ...     [1.   , 0.  , 0.   ],
           ...     [0.25 , 0.5 , 0.25 ],
           ...     [0.375, 0.25, 0.375],
           ... ])
           >>> points = triangle.evaluate_barycentric_multi(param_vals)
           >>> points
           array([[-1.75 , 0. , 0.25   , -0.625   ],
                  [ 1.75 , 0. , 1.0625 ,  1.046875]])

        .. testcleanup:: triangle-eval-multi2

           import make_images
           make_images.triangle_evaluate_barycentric_multi(triangle, points)

        Args:
            param_vals (numpy.ndarray): Array of parameter values (as a
                ``N x 3`` array).
            _verify (Optional[bool]): Indicates if the coordinates should be
                verified. See :meth:`evaluate_barycentric`. Defaults to
                :data:`True`. Will also double check that ``param_vals``
                is the right shape.

        Returns:
            numpy.ndarray: The points on the triangle.

        Raises:
            ValueError: If ``param_vals`` is not a 2D array and
                ``_verify=True``.
        """
        if _verify:
            if param_vals.ndim != 2:
                raise ValueError("Parameter values must be 2D array")

            for lambda1, lambda2, lambda3 in param_vals:
                self._verify_barycentric(lambda1, lambda2, lambda3)
        return _triangle_helpers.evaluate_barycentric_multi(
            self._nodes, self._degree, param_vals, self._dimension
        )

    @staticmethod
    def _verify_cartesian(s, t):
        """Verifies that a point is in the reference triangle.

        I.e., checks that they sum to <= one and are each non-negative.

        Args:
            s (float): Parameter along the reference triangle.
            t (float): Parameter along the reference triangle.

        Raises:
            ValueError: If the point lies outside the reference triangle.
        """
        if s < 0.0 or t < 0.0 or s + t > 1.0:
            raise ValueError("Point lies outside reference triangle", s, t)

    def evaluate_cartesian(self, s, t, _verify=True):
        r"""Compute a point on the triangle.

        Evaluates :math:`B\left(1 - s - t, s, t\right)` by calling
        :meth:`evaluate_barycentric`:

        This method acts as a (partial) inverse to :meth:`locate`.

        .. testsetup:: triangle-cartesian

           import numpy as np
           import bezier

        .. doctest:: triangle-cartesian
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5, 1.0  , 0.0, 0.5, 0.25],
           ...     [0.0, 0.5, 0.625, 0.5, 0.5, 1.0 ],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> point = triangle.evaluate_cartesian(0.125, 0.375)
           >>> point
           array([[0.16015625],
                  [0.44726562]])
           >>> triangle.evaluate_barycentric(0.5, 0.125, 0.375)
           array([[0.16015625],
                  [0.44726562]])

        Args:
            s (float): Parameter along the reference triangle.
            t (float): Parameter along the reference triangle.
            _verify (Optional[bool]): Indicates if the coordinates should be
                verified inside of the reference triangle. Defaults to
                :data:`True`.

        Returns:
            numpy.ndarray: The point on the triangle (as a two dimensional
            NumPy array).
        """
        if _verify:
            self._verify_cartesian(s, t)
        return _triangle_helpers.evaluate_barycentric(
            self._nodes, self._degree, 1.0 - s - t, s, t
        )

    def evaluate_cartesian_multi(self, param_vals, _verify=True):
        r"""Compute multiple points on the triangle.

        Assumes ``param_vals`` has two columns of Cartesian coordinates.
        See :meth:`evaluate_cartesian` for more details on how each row of
        parameter values is evaluated.

        .. image:: ../../images/triangle_evaluate_cartesian_multi.png
           :align: center

        .. doctest:: triangle-eval-multi1
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 2.0, -3.0],
           ...     [0.0, 1.0,  2.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=1)
           >>> triangle
           <Triangle (degree=1, dimension=2)>
           >>> param_vals = np.asfortranarray([
           ...     [0.0  , 0.0  ],
           ...     [0.125, 0.625],
           ...     [0.5  , 0.5  ],
           ... ])
           >>> points = triangle.evaluate_cartesian_multi(param_vals)
           >>> points
           array([[ 0. , -1.625, -0.5 ],
                  [ 0. ,  1.375,  1.5 ]])

        .. testcleanup:: triangle-eval-multi1

           import make_images
           make_images.triangle_evaluate_cartesian_multi(triangle, points)

        Args:
            param_vals (numpy.ndarray): Array of parameter values (as a
                ``N x 2`` array).
            _verify (Optional[bool]): Indicates if the coordinates should be
                verified. See :meth:`evaluate_cartesian`. Defaults to
                :data:`True`. Will also double check that ``param_vals``
                is the right shape.

        Returns:
            numpy.ndarray: The points on the triangle.

        Raises:
            ValueError: If ``param_vals`` is not a 2D array and
                ``_verify=True``.
        """
        if _verify:
            if param_vals.ndim != 2:
                raise ValueError("Parameter values must be 2D array")

            for s, t in param_vals:
                self._verify_cartesian(s, t)
        return _triangle_helpers.evaluate_cartesian_multi(
            self._nodes, self._degree, param_vals, self._dimension
        )

    def plot(self, pts_per_edge, color=None, ax=None, with_nodes=False):
        """Plot the current triangle.

        Args:
            pts_per_edge (int): Number of points to plot per edge.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.
            with_nodes (Optional[bool]): Determines if the control points
                should be added to the plot. Off by default.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.

        Raises:
            NotImplementedError: If the triangle's dimension is not ``2``.
        """
        if self._dimension != 2:
            raise NotImplementedError(
                "2D is the only supported dimension",
                "Current dimension",
                self._dimension,
            )

        if ax is None:
            ax = _plot_helpers.new_axis()
        _plot_helpers.add_patch(ax, color, pts_per_edge, *self._get_edges())
        if with_nodes:
            ax.plot(
                self._nodes[0, :],
                self._nodes[1, :],
                color="black",
                marker="o",
                linestyle="None",
            )
        return ax

    def subdivide(self):
        r"""Split the triangle into four sub-triangles.

        Does so by taking the unit triangle (i.e. the domain
        of the triangle) and splitting it into four sub-triangles

        .. image:: ../../images/triangle_subdivide1.png
           :align: center

        Then the triangle is re-parameterized via the map to / from the
        given sub-triangles and the unit triangle.

        For example, when a degree two triangle is subdivided:

        .. image:: ../../images/triangle_subdivide2.png
           :align: center

        .. doctest:: triangle-subdivide
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [-1.0, 0.5, 2.0, 0.25, 2.0, 0.0],
           ...     [ 0.0, 0.5, 0.0, 1.75, 3.0, 4.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> _, sub_triangle_b, _, _ = triangle.subdivide()
           >>> sub_triangle_b
           <Triangle (degree=2, dimension=2)>
           >>> sub_triangle_b.nodes
           array([[ 1.5 ,  0.6875, -0.125 , 1.1875, 0.4375, 0.5  ],
                  [ 2.5 ,  2.3125,  1.875 , 1.3125, 1.3125, 0.25 ]])

        .. testcleanup:: triangle-subdivide

           import make_images
           make_images.triangle_subdivide1()
           make_images.triangle_subdivide2(triangle, sub_triangle_b)

        Returns:
            Tuple[Triangle, Triangle, Triangle, Triangle]: The lower left,
            central, lower right and upper left sub-triangles (in that order).
        """
        (
            nodes_a,
            nodes_b,
            nodes_c,
            nodes_d,
        ) = _triangle_helpers.subdivide_nodes(self._nodes, self._degree)
        return (
            Triangle(nodes_a, self._degree, copy=False, verify=False),
            Triangle(nodes_b, self._degree, copy=False, verify=False),
            Triangle(nodes_c, self._degree, copy=False, verify=False),
            Triangle(nodes_d, self._degree, copy=False, verify=False),
        )

    def _compute_valid(self):
        r"""Determines if the current triangle is "valid".

        Does this by checking if the Jacobian of the map from the
        reference triangle is everywhere positive.

        Returns:
            bool: Flag indicating if the current triangle is valid.

        Raises:
            NotImplementedError: If the triangle is in a dimension other
                than :math:`\mathbf{R}^2`.
            .UnsupportedDegree: If the degree is not 1, 2 or 3.
        """
        if self._dimension != 2:
            raise NotImplementedError("Validity check only implemented in R^2")

        poly_sign = None
        if self._degree == 1:
            # In the linear case, we are only invalid if the points
            # are collinear.
            first_deriv = self._nodes[:, 1:] - self._nodes[:, :-1]
            # pylint: disable=assignment-from-no-return
            poly_sign = _SIGN(np.linalg.det(first_deriv))
            # pylint: enable=assignment-from-no-return
        elif self._degree == 2:
            bernstein = _py_triangle_helpers.quadratic_jacobian_polynomial(
                self._nodes
            )
            poly_sign = _py_triangle_helpers.polynomial_sign(bernstein, 2)
        elif self._degree == 3:
            bernstein = _py_triangle_helpers.cubic_jacobian_polynomial(
                self._nodes
            )
            poly_sign = _py_triangle_helpers.polynomial_sign(bernstein, 4)
        else:
            raise _py_helpers.UnsupportedDegree(
                self._degree, supported=(1, 2, 3)
            )

        return poly_sign == 1

    @property
    def is_valid(self):
        """bool: Flag indicating if the triangle is "valid".

        Here, "valid" means there are no self-intersections or
        singularities and the edges are oriented with the interior
        (i.e. a 90 degree rotation of the tangent vector to the left is
        the interior).

        This checks if the Jacobian of the map from the reference
        triangle is everywhere positive. For example, a linear "triangle"
        with collinear points is invalid:

        .. image:: ../../images/triangle_is_valid1.png
           :align: center

        .. doctest:: triangle-is-valid1

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.0, 2.0],
           ...     [0.0, 1.0, 2.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=1)
           >>> triangle.is_valid
           False

        .. testcleanup:: triangle-is-valid1

           import make_images
           make_images.triangle_is_valid1(triangle)

        while a quadratic triangle with one straight side:

        .. image:: ../../images/triangle_is_valid2.png
           :align: center

        .. doctest:: triangle-is-valid2

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5  , 1.0, -0.125, 0.5, 0.0],
           ...     [0.0, 0.125, 0.0,  0.5  , 0.5, 1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> triangle.is_valid
           True

        .. testcleanup:: triangle-is-valid2

           import make_images
           make_images.triangle_is_valid2(triangle)

        though not all higher degree triangles are valid:

        .. image:: ../../images/triangle_is_valid3.png
           :align: center

        .. doctest:: triangle-is-valid3

           >>> nodes = np.asfortranarray([
           ...     [1.0, 0.0, 1.0, 0.0, 0.0, 0.0],
           ...     [0.0, 0.0, 1.0, 0.0, 0.0, 1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> triangle.is_valid
           False

        .. testcleanup:: triangle-is-valid3

           import make_images
           make_images.triangle_is_valid3(triangle)
        """
        return self._compute_valid()

    @property
    def __dict__(self):
        """dict: Dictionary of current triangle's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {
            "_dimension": self._dimension,
            "_nodes": self._nodes,
            "_degree": self._degree,
            "_edges": self._edges,
        }

    def locate(self, point, _verify=True):
        r"""Find a point on the current triangle.

        Solves for :math:`s` and :math:`t` in :math:`B(s, t) = p`.

        This method acts as a (partial) inverse to :meth:`evaluate_cartesian`.

        .. warning::

           A unique solution is only guaranteed if the current triangle is
           valid. This code assumes a valid triangle, but doesn't check.

        .. image:: ../../images/triangle_locate.png
           :align: center

        .. doctest:: triangle-locate

           >>> nodes = np.asfortranarray([
           ...     [0.0,  0.5 , 1.0, 0.25, 0.75, 0.0],
           ...     [0.0, -0.25, 0.0, 0.5 , 0.75, 1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> point = np.asfortranarray([
           ...     [0.59375],
           ...     [0.25   ],
           ... ])
           >>> s, t = triangle.locate(point)
           >>> s
           0.5
           >>> t
           0.25

        .. testcleanup:: triangle-locate

           import make_images
           make_images.triangle_locate(triangle, point)

        Args:
            point (numpy.ndarray): A (``D x 1``) point on the triangle,
                where :math:`D` is the dimension of the triangle.
            _verify (Optional[bool]): Indicates if extra caution should be
                used to verify assumptions about the inputs. Can be
                disabled to speed up execution time. Defaults to :data:`True`.

        Returns:
            Optional[Tuple[float, float]]: The :math:`s` and :math:`t`
            values corresponding to ``point`` or :data:`None` if the point
            is not on the triangle.

        Raises:
            NotImplementedError: If the triangle isn't in :math:`\mathbf{R}^2`.
            ValueError: If the dimension of the ``point`` doesn't match the
                dimension of the current triangle.
        """
        if _verify:
            if self._dimension != 2:
                raise NotImplementedError("Only 2D triangles supported.")

            if point.shape != (self._dimension, 1):
                point_dimensions = " x ".join(
                    str(dimension) for dimension in point.shape
                )
                msg = _LOCATE_ERROR_TEMPLATE.format(
                    self._dimension, self._dimension, point, point_dimensions
                )
                raise ValueError(msg)

        return _triangle_intersection.locate_point(
            self._nodes, self._degree, point[0, 0], point[1, 0]
        )

    def intersect(self, other, strategy=_STRATEGY.GEOMETRIC, _verify=True):
        """Find the common intersection with another triangle.

        Args:
            other (Triangle): Other triangle to intersect with.
            strategy (Optional[~bezier.curve.IntersectionStrategy]): The
                intersection algorithm to use. Defaults to geometric.
            _verify (Optional[bool]): Indicates if extra caution should be
                used to verify assumptions about the algorithm as it
                proceeds. Can be disabled to speed up execution time.
                Defaults to :data:`True`.

        Returns:
            List[Union[~bezier.curved_polygon.CurvedPolygon, \
            ~bezier.triangle.Triangle]]: List of intersections (possibly
            empty).

        Raises:
            TypeError: If ``other`` is not a triangle (and ``_verify=True``).
            NotImplementedError: If at least one of the triangles
                isn't two-dimensional (and ``_verify=True``).
            ValueError: If ``strategy`` is not a valid
                :class:`.IntersectionStrategy`.
        """
        if _verify:
            if not isinstance(other, Triangle):
                raise TypeError(
                    "Can only intersect with another triangle",
                    "Received",
                    other,
                )

            if self._dimension != 2 or other._dimension != 2:
                raise NotImplementedError(
                    "Intersection only implemented in 2D"
                )

        if strategy == _STRATEGY.GEOMETRIC:
            do_intersect = _triangle_intersection.geometric_intersect
        elif strategy == _STRATEGY.ALGEBRAIC:
            do_intersect = _py_triangle_intersection.algebraic_intersect
        else:
            raise ValueError("Unexpected strategy.", strategy)

        edge_infos, contained, all_edge_nodes = do_intersect(
            self._nodes, self._degree, other._nodes, other._degree, _verify
        )
        if edge_infos is None:
            if contained:
                return [self]

            else:
                return [other]

        else:
            return [
                _make_intersection(edge_info, all_edge_nodes)
                for edge_info in edge_infos
            ]

    def elevate(self):
        r"""Return a degree-elevated version of the current triangle.

        Does this by converting the current nodes
        :math:`\left\{v_{i, j, k}\right\}_{i + j + k = d}` to new nodes
        :math:`\left\{w_{i, j, k}\right\}_{i + j + k = d + 1}`. Does so
        by re-writing

        .. math::

           E\left(\lambda_1, \lambda_2, \lambda_3\right) =
               \left(\lambda_1 + \lambda_2 + \lambda_3\right)
               B\left(\lambda_1, \lambda_2, \lambda_3\right) =
               \sum_{i + j + k = d + 1} \binom{d + 1}{i \, j \, k}
               \lambda_1^i \lambda_2^j \lambda_3^k \cdot w_{i, j, k}

        In this form, we must have

        .. math::

           \begin{align*}
           \binom{d + 1}{i \, j \, k} \cdot w_{i, j, k} &=
               \binom{d}{i - 1 \, j \, k} \cdot v_{i - 1, j, k} +
               \binom{d}{i \, j - 1 \, k} \cdot v_{i, j - 1, k} +
               \binom{d}{i \, j \, k - 1} \cdot v_{i, j, k - 1} \\
           \Longleftrightarrow (d + 1) \cdot w_{i, j, k} &=
               i \cdot v_{i - 1, j, k} + j \cdot v_{i, j - 1, k} +
               k \cdot v_{i, j, k - 1}
           \end{align*}

        where we define, for example, :math:`v_{i, j, k - 1} = 0`
        if :math:`k = 0`.

        .. image:: ../../images/triangle_elevate.png
           :align: center

        .. doctest:: triangle-elevate
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.asfortranarray([
           ...     [0.0, 1.5, 3.0, 0.75, 2.25, 0.0],
           ...     [0.0, 0.0, 0.0, 1.5 , 2.25, 3.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> elevated = triangle.elevate()
           >>> elevated
           <Triangle (degree=3, dimension=2)>
           >>> elevated.nodes
           array([[0. , 1. , 2. , 3. , 0.5 , 1.5 , 2.5 , 0.5 , 1.5 , 0. ],
                  [0. , 0. , 0. , 0. , 1.  , 1.25, 1.5 , 2.  , 2.5 , 3. ]])

        .. testcleanup:: triangle-elevate

           import make_images
           make_images.triangle_elevate(triangle, elevated)

        Returns:
            Triangle: The degree-elevated triangle.
        """
        _, num_nodes = self._nodes.shape
        # (d + 1)(d + 2)/2 --> (d + 2)(d + 3)/2
        num_new = num_nodes + self._degree + 2
        new_nodes = np.zeros((self._dimension, num_new), order="F")
        # NOTE: We start from the index triples (i, j, k) for the current
        #       nodes and map them onto (i + 1, j, k), etc. This index
        #       tracking is also done in :func:`.de_casteljau_one_round`.
        index = 0
        # parent_i1 = index + k
        # parent_i2 = index + k + 1
        # parent_i3 = index + degree + 2
        parent_i1 = 0
        parent_i2 = 1
        parent_i3 = self._degree + 2
        for k in range(self._degree + 1):
            for j in range(self._degree + 1 - k):
                i = self._degree - j - k
                new_nodes[:, parent_i1] += (i + 1) * self._nodes[:, index]
                new_nodes[:, parent_i2] += (j + 1) * self._nodes[:, index]
                new_nodes[:, parent_i3] += (k + 1) * self._nodes[:, index]
                # Update all the indices.
                parent_i1 += 1
                parent_i2 += 1
                parent_i3 += 1
                index += 1
            # Update the indices that depend on k.
            parent_i1 += 1
            parent_i2 += 1
        # Hold off on division until the end, to (attempt to) avoid round-off.
        denominator = self._degree + 1.0
        new_nodes /= denominator
        return Triangle(new_nodes, self._degree + 1, copy=False, verify=False)

    # Return type doc appears missing to Pylint because of the use of the
    # :class:`sympy.Matrix ...` aliases.
    # pylint: disable=missing-return-type-doc
    def to_symbolic(self):
        """Convert to a SymPy matrix representing :math:`B(s, t)`.

        .. note::

           This method requires SymPy.

        .. doctest:: triangle-to-symbolic

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5, 1.0, -0.5, 0.0, -1.0],
           ...     [0.0, 0.0, 1.0,  0.0, 0.0,  0.0],
           ...     [0.0, 0.0, 0.0,  0.0, 0.0,  1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> triangle.to_symbolic()
           Matrix([
           [s - t],
           [ s**2],
           [ t**2]])

        Returns:
            :class:`sympy.Matrix <sympy.matrices.dense.MutableDenseMatrix>`:
            The triangle :math:`B(s, t)`.
        """
        _, _, b_polynomial = _symbolic.triangle_as_polynomial(
            self._nodes, self._degree
        )
        return b_polynomial

    def implicitize(self):
        r"""Implicitize the triangle .

        .. note::

           This method requires SymPy.

        .. doctest:: triangle-implicitize

           >>> nodes = np.asfortranarray([
           ...     [0.0, 0.5, 1.0, -0.5, 0.0, -1.0],
           ...     [0.0, 0.0, 1.0,  0.0, 0.0,  0.0],
           ...     [0.0, 0.0, 0.0,  0.0, 0.0,  1.0],
           ... ])
           >>> triangle = bezier.Triangle(nodes, degree=2)
           >>> triangle.implicitize()
           (x**4 - 2*x**2*y - 2*x**2*z + y**2 - 2*y*z + z**2)**2

        Returns:
            :class:`sympy.Expr <sympy.core.expr.Expr>`: The function that
            defines the triangle in :math:`\mathbf{R}^3` via
            :math:`f(x, y, z) = 0`.

        Raises:
            ValueError: If the triangle's dimension is not ``3``.
        """
        if self._dimension != 3:
            raise ValueError(
                "Only a spatial (3D) triangle can be implicitized",
                "Current dimension",
                self._dimension,
            )

        return _symbolic.implicitize_triangle(self._nodes, self._degree)

    # pylint: enable=missing-return-type-doc


def _make_intersection(edge_info, all_edge_nodes):
    """Convert a description of edges into a curved polygon.

    .. note::

       This is a helper used only by :meth:`.Triangle.intersect`.

    Args:
        edge_info (Tuple[Tuple[int, float, float], ...]): Information
            describing each edge in the curved polygon by indicating which
            triangle / edge on the triangle and then start and end parameters
            along that edge. (See :func:`.ends_to_curve`.)
        all_edge_nodes (Tuple[numpy.ndarray, ...]): The nodes of three edges
            of the first triangle being intersected followed by the nodes of
            the three edges of the second.

    Returns:
        .CurvedPolygon: The intersection corresponding to ``edge_info``.
    """
    edges = []
    for index, start, end in edge_info:
        nodes = all_edge_nodes[index]
        new_nodes = _curve_helpers.specialize_curve(nodes, start, end)
        degree = new_nodes.shape[1] - 1
        edge = _curve_mod.Curve(new_nodes, degree, copy=False, verify=False)
        edges.append(edge)
    return curved_polygon.CurvedPolygon(
        *edges, metadata=edge_info, _verify=False
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for :mod:`bezier.triangle`.

The functions provided by this module have a Cython speedup with the
exact same interface which calls out to a Fortran implementation. The speedup
will be used if the extension can be built.
"""

import atexit

from bezier import _py_triangle_intersection

try:
    from bezier import _speedup
except ImportError:  # pragma: NO COVER
    _speedup = None


# pylint: disable=invalid-name
if _speedup is None:  # pragma: NO COVER
    newton_refine = _py_triangle_intersection.newton_refine
    locate_point = _py_triangle_intersection.locate_point
    geometric_intersect = _py_triangle_intersection.geometric_intersect
else:
    newton_refine = _speedup.newton_refine_triangle
    locate_point = _speedup.locate_point_triangle
    geometric_intersect = _speedup.triangle_intersections
    atexit.register(_speedup.free_triangle_intersections_workspace)
# pylint: enable=invalid-name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pure Python helper methods for :mod:`bezier.triangle`."""

import collections
import itertools

import numpy as np

from bezier import _algebraic_intersection
from bezier import _py_geometric_intersection
from bezier import _py_helpers
from bezier import _py_intersection_helpers
from bezier import _py_triangle_helpers


MAX_LOCATE_SUBDIVISIONS = 20
LOCATE_EPS = 0.5 ** 47
INTERSECTION_T = _py_geometric_intersection.BoxIntersectionType.INTERSECTION
CLASSIFICATION_T = _py_intersection_helpers.IntersectionClassification
UNUSED_T = CLASSIFICATION_T.COINCIDENT_UNUSED
ACCEPTABLE_CLASSIFICATIONS = (
    CLASSIFICATION_T.FIRST,
    CLASSIFICATION_T.SECOND,
    CLASSIFICATION_T.COINCIDENT,
)
TANGENT_CLASSIFICATIONS = (
    CLASSIFICATION_T.TANGENT_FIRST,
    CLASSIFICATION_T.TANGENT_SECOND,
)
BAD_SEGMENT_PARAMS = (
    "Segment start should be strictly less than end and both "
    "should be in [0, 1]."
)
SEGMENTS_SAME_EDGE = "Consecutive segments lie on the same edge."


def newton_refine_solve(jac_both, x_val, triangle_x, y_val, triangle_y):
    r"""Helper for :func:`newton_refine`.

    We have a system:

    .. code-block:: rest

       [A C][ds] = [E]
       [B D][dt]   [F]

    This is not a typo, ``A->B->C->D`` matches the data in ``jac_both``.
    We solve directly rather than using a linear algebra utility:

    .. code-block:: rest

       ds = (D E - C F) / (A D - B C)
       dt = (A F - B E) / (A D - B C)

    Args:
        jac_both (numpy.ndarray): A ``4 x 1`` matrix of entries in a Jacobian.
        x_val (float): An ``x``-value we are trying to reach.
        triangle_x (float): The actual ``x``-value we are currently at.
        y_val (float): An ``y``-value we are trying to reach.
        triangle_y (float): The actual ``x``-value we are currently at.

    Returns:
        Tuple[float, float]: The pair of values the solve the
         linear system.
    """
    a_val, b_val, c_val, d_val = jac_both[:, 0]
    #       and
    e_val = x_val - triangle_x
    f_val = y_val - triangle_y
    # Now solve:
    denom = a_val * d_val - b_val * c_val
    delta_s = (d_val * e_val - c_val * f_val) / denom
    delta_t = (a_val * f_val - b_val * e_val) / denom
    return delta_s, delta_t


def newton_refine(nodes, degree, x_val, y_val, s, t):
    r"""Refine a solution to :math:`B(s, t) = p` using Newton's method.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Computes updates via

    .. math::

       \left[\begin{array}{c}
           0 \\ 0 \end{array}\right] \approx
           \left(B\left(s_{\ast}, t_{\ast}\right) -
           \left[\begin{array}{c} x \\ y \end{array}\right]\right) +
           \left[\begin{array}{c c}
               B_s\left(s_{\ast}, t_{\ast}\right) &
               B_t\left(s_{\ast}, t_{\ast}\right) \end{array}\right]
           \left[\begin{array}{c}
               \Delta s \\ \Delta t \end{array}\right]

    For example, (with weights
    :math:`\lambda_1 = 1 - s - t, \lambda_2 = s, \lambda_3 = t`)
    consider the triangle

    .. math::

       B(s, t) =
           \left[\begin{array}{c} 0 \\ 0 \end{array}\right] \lambda_1^2 +
           \left[\begin{array}{c} 1 \\ 0 \end{array}\right]
               2 \lambda_1 \lambda_2 +
           \left[\begin{array}{c} 2 \\ 0 \end{array}\right] \lambda_2^2 +
           \left[\begin{array}{c} 2 \\ 1 \end{array}\right]
               2 \lambda_1 \lambda_3 +
           \left[\begin{array}{c} 2 \\ 2 \end{array}\right]
               2 \lambda_2 \lambda_1 +
           \left[\begin{array}{c} 0 \\ 2 \end{array}\right] \lambda_3^2

    and the point
    :math:`B\left(\frac{1}{4}, \frac{1}{2}\right) =
    \frac{1}{4} \left[\begin{array}{c} 5 \\ 5 \end{array}\right]`.

    Starting from the **wrong** point
    :math:`s = \frac{1}{2}, t = \frac{1}{4}`, we have

    .. math::

       \begin{align*}
       \left[\begin{array}{c} x \\ y \end{array}\right] -
          B\left(\frac{1}{2}, \frac{1}{4}\right) &= \frac{1}{4}
          \left[\begin{array}{c} -1 \\ 2 \end{array}\right] \\
       DB\left(\frac{1}{2}, \frac{1}{4}\right) &= \frac{1}{2}
           \left[\begin{array}{c c} 3 & 2 \\ 1 & 6 \end{array}\right] \\
       \Longrightarrow \left[\begin{array}{c}
           \Delta s \\ \Delta t \end{array}\right] &= \frac{1}{32}
           \left[\begin{array}{c} -10 \\ 7 \end{array}\right]
       \end{align*}

    .. image:: ../images/newton_refine_triangle.png
       :align: center

    .. testsetup:: newton-refine-triangle

       import numpy as np
       import bezier
       from bezier._py_triangle_intersection import newton_refine

    .. doctest:: newton-refine-triangle

       >>> nodes = np.asfortranarray([
       ...     [0.0, 1.0, 2.0, 2.0, 2.0, 0.0],
       ...     [0.0, 0.0, 0.0, 1.0, 2.0, 2.0],
       ... ])
       >>> triangle = bezier.Triangle(nodes, degree=2)
       >>> triangle.is_valid
       True
       >>> (x_val,), (y_val,) = triangle.evaluate_cartesian(0.25, 0.5)
       >>> x_val, y_val
       (1.25, 1.25)
       >>> s, t = 0.5, 0.25
       >>> new_s, new_t = newton_refine(nodes, 2, x_val, y_val, s, t)
       >>> 32 * (new_s - s)
       -10.0
       >>> 32 * (new_t - t)
       7.0

    .. testcleanup:: newton-refine-triangle

       import make_images
       make_images.newton_refine_triangle(
           triangle, x_val, y_val, s, t, new_s, new_t)

    Args:
        nodes (numpy.ndarray): Array of nodes in a triangle.
        degree (int): The degree of the triangle.
        x_val (float): The :math:`x`-coordinate of a point
            on the triangle.
        y_val (float): The :math:`y`-coordinate of a point
            on the triangle.
        s (float): Approximate :math:`s`-value to be refined.
        t (float): Approximate :math:`t`-value to be refined.

    Returns:
        Tuple[float, float]: The refined :math:`s` and :math:`t` values.
    """
    lambda1 = 1.0 - s - t
    (triangle_x,), (triangle_y,) = _py_triangle_helpers.evaluate_barycentric(
        nodes, degree, lambda1, s, t
    )
    if triangle_x == x_val and triangle_y == y_val:
        # No refinement is needed.
        return s, t

    # NOTE: This function assumes ``dimension==2`` (i.e. since ``x, y``).
    jac_nodes = _py_triangle_helpers.jacobian_both(nodes, degree, 2)
    # The degree of the jacobian is one less.
    jac_both = _py_triangle_helpers.evaluate_barycentric(
        jac_nodes, degree - 1, lambda1, s, t
    )
    # The first row of the jacobian matrix is B_s (i.e. the
    # top-most values in ``jac_both``).
    delta_s, delta_t = newton_refine_solve(
        jac_both, x_val, triangle_x, y_val, triangle_y
    )
    return s + delta_s, t + delta_t


def update_locate_candidates(candidate, next_candidates, x_val, y_val, degree):
    """Update list of candidate triangles during geometric search for a point.

    .. note::

       This is used **only** as a helper for :func:`locate_point`.

    Checks if the point ``(x_val, y_val)`` is contained in the ``candidate``
    triangle. If not, this function does nothing. If the point is contaned,
    the four subdivided triangles from ``candidate`` are added to
    ``next_candidates``.

    Args:
        candidate (Tuple[float, float, float, numpy.ndarray]): A 4-tuple
            describing a triangle and its centroid / width. Contains

            * Three times centroid ``x``-value
            * Three times centroid ``y``-value
            * "Width" of parameter space for the triangle
            * Control points for the triangle
        next_candidates (list): List of "candidate" sub-triangles that may
            contain the point being located.
        x_val (float): The ``x``-coordinate being located.
        y_val (float): The ``y``-coordinate being located.
        degree (int): The degree of the triangle.
    """
    centroid_x, centroid_y, width, candidate_nodes = candidate
    point = np.asfortranarray([x_val, y_val])
    if not _py_helpers.contains_nd(candidate_nodes, point):
        return

    nodes_a, nodes_b, nodes_c, nodes_d = _py_triangle_helpers.subdivide_nodes(
        candidate_nodes, degree
    )
    half_width = 0.5 * width
    next_candidates.extend(
        (
            (
                centroid_x - half_width,
                centroid_y - half_width,
                half_width,
                nodes_a,
            ),
            (centroid_x, centroid_y, -half_width, nodes_b),
            (centroid_x + width, centroid_y - half_width, half_width, nodes_c),
            (centroid_x - half_width, centroid_y + width, half_width, nodes_d),
        )
    )


def mean_centroid(candidates):
    """Take the mean of all centroids in set of reference triangles.

    .. note::

       This is used **only** as a helper for :func:`locate_point`.

    Args:
        candidates (List[Tuple[float, float, float, numpy.ndarray]): List of
            4-tuples, each of which has been produced by :func:`locate_point`.
            Each 4-tuple contains

            * Three times centroid ``x``-value
            * Three times centroid ``y``-value
            * "Width" of a parameter space for a triangle
            * Control points for a triangle

            We only use the first two values, which are triple the desired
            value so that we can put off division by three until summing in
            our average. We don't use the other two values, they are just an
            artifact of the way ``candidates`` is constructed by the caller.

    Returns:
        Tuple[float, float]: The mean of all centroids.
    """
    sum_x = 0.0
    sum_y = 0.0
    for centroid_x, centroid_y, _, _ in candidates:
        sum_x += centroid_x
        sum_y += centroid_y
    denom = 3.0 * len(candidates)
    return sum_x / denom, sum_y / denom


def locate_point(nodes, degree, x_val, y_val):
    r"""Locate a point on a triangle.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Does so by recursively subdividing the triangle and rejecting
    sub-triangles with bounding boxes that don't contain the point.
    After the sub-triangles are sufficiently small, uses Newton's
    method to narrow in on the pre-image of the point.

    Args:
        nodes (numpy.ndarray): Control points for B |eacute| zier triangle
            (assumed to be two-dimensional).
        degree (int): The degree of the triangle.
        x_val (float): The :math:`x`-coordinate of a point
            on the triangle.
        y_val (float): The :math:`y`-coordinate of a point
            on the triangle.

    Returns:
        Optional[Tuple[float, float]]: The :math:`s` and :math:`t`
        values corresponding to ``x_val`` and ``y_val`` or
        :data:`None` if the point is not on the ``triangle``.
    """
    # We track the centroid rather than base_x/base_y/width (by storing triple
    # the centroid -- to avoid division by three until needed). We also need
    # to track the width (or rather, just the sign of the width).
    candidates = [(1.0, 1.0, 1.0, nodes)]
    for _ in range(MAX_LOCATE_SUBDIVISIONS + 1):
        next_candidates = []
        for candidate in candidates:
            update_locate_candidates(
                candidate, next_candidates, x_val, y_val, degree
            )
        candidates = next_candidates
    if not candidates:
        return None

    # We take the average of all centroids from the candidates
    # that may contain the point.
    s_approx, t_approx = mean_centroid(candidates)
    s, t = newton_refine(nodes, degree, x_val, y_val, s_approx, t_approx)
    actual = _py_triangle_helpers.evaluate_barycentric(
        nodes, degree, 1.0 - s - t, s, t
    )
    expected = np.asfortranarray([x_val, y_val])
    if not _py_helpers.vector_close(
        actual.ravel(order="F"), expected, eps=LOCATE_EPS
    ):
        s, t = newton_refine(nodes, degree, x_val, y_val, s, t)
    return s, t


def same_intersection(intersection1, intersection2, wiggle=0.5 ** 40):
    """Check if two intersections are close to machine precision.

    .. note::

       This is a helper used only by :func:`verify_duplicates`, which in turn
       is only used by :func:`generic_intersect`.

    Args:
        intersection1 (.Intersection): The first intersection.
        intersection2 (.Intersection): The second intersection.
        wiggle (Optional[float]): The amount of relative error allowed
            in parameter values.

    Returns:
        bool: Indicates if the two intersections are the same to
        machine precision.
    """
    if intersection1.index_first != intersection2.index_first:
        return False

    if intersection1.index_second != intersection2.index_second:
        return False

    return np.allclose(
        [intersection1.s, intersection1.t],
        [intersection2.s, intersection2.t],
        atol=0.0,
        rtol=wiggle,
    )


def verify_duplicates(duplicates, uniques):
    """Verify that a set of intersections had expected duplicates.

    .. note::

       This is a helper used only by :func:`generic_intersect`.

    Args:
        duplicates (List[.Intersection]): List of intersections
            corresponding to duplicates that were filtered out.
        uniques (List[.Intersection]): List of "final" intersections
            with duplicates filtered out.

    Raises:
        ValueError: If the ``uniques`` are not actually all unique.
        ValueError: If one of the ``duplicates`` does not correspond to
            an intersection in ``uniques``.
        ValueError: If a duplicate occurs only once but does not have
            exactly one of ``s`` and ``t`` equal to ``0.0``.
        ValueError: If a duplicate occurs three times but does not have
            exactly both ``s == t == 0.0``.
        ValueError: If a duplicate occurs a number other than one or three
            times.
    """
    for uniq1, uniq2 in itertools.combinations(uniques, 2):
        if same_intersection(uniq1, uniq2):
            raise ValueError("Non-unique intersection")

    counter = collections.Counter()
    for dupe in duplicates:
        matches = []
        for index, uniq in enumerate(uniques):
            if same_intersection(dupe, uniq):
                matches.append(index)
        if len(matches) != 1:
            raise ValueError("Duplicate not among uniques", dupe)

        matched = matches[0]
        counter[matched] += 1
    for index, count in counter.items():
        uniq = uniques[index]
        if count == 1:
            if (uniq.s, uniq.t).count(0.0) != 1:
                raise ValueError("Count == 1 should be a single corner", uniq)

        elif count == 3:
            if (uniq.s, uniq.t) != (0.0, 0.0):
                raise ValueError("Count == 3 should be a double corner", uniq)

        else:
            raise ValueError("Unexpected duplicate count", count)


def verify_edge_segments(edge_infos):
    """Verify that the edge segments in an intersection are valid.

    .. note::

       This is a helper used only by :func:`generic_intersect`.

    Args:
        edge_infos (Optional[list]): List of "edge info" lists. Each list
            represents a curved polygon and contains 3-tuples of edge index,
            start and end (see the output of :func:`ends_to_curve`).

    Raises:
        ValueError: If two consecutive edge segments lie on the same edge
            index.
        ValueError: If the start and end parameter are "invalid" (they should
            be between 0 and 1 and start should be strictly less than end).
    """
    if edge_infos is None:
        return

    for edge_info in edge_infos:
        num_segments = len(edge_info)
        for index in range(-1, num_segments - 1):
            index1, start1, end1 = edge_info[index]
            # First, verify the start and end parameters for the current
            # segment.
            if not 0.0 <= start1 < end1 <= 1.0:
                raise ValueError(BAD_SEGMENT_PARAMS, edge_info[index])

            # Then, verify that the indices are not the same.
            index2, _, _ = edge_info[index + 1]
            if index1 == index2:
                raise ValueError(
                    SEGMENTS_SAME_EDGE, edge_info[index], edge_info[index + 1]
                )


def add_edge_end_unused(intersection, duplicates, intersections):
    """Add intersection that is ``COINCIDENT_UNUSED`` but on an edge end.

    This is a helper for :func:`~._py_triangle_intersection.add_intersection`.
    It assumes that

    * ``intersection`` will have at least one of ``s == 0.0`` or ``t == 0.0``
    * A "misclassified" intersection in ``intersections`` that matches
      ``intersection`` will be the "same" if it matches both ``index_first``
      and ``index_second`` and if it matches the start index exactly

    Args:
        intersection (.Intersection): An intersection to be added.
        duplicates (List[.Intersection]): List of duplicate intersections.
        intersections (List[.Intersection]): List of "accepted" (i.e.
            non-duplicate) intersections.
    """
    found = None
    for other in intersections:
        if (
            intersection.index_first == other.index_first
            and intersection.index_second == other.index_second
        ):
            if intersection.s == 0.0 and other.s == 0.0:
                found = other
                break

            if intersection.t == 0.0 and other.t == 0.0:
                found = other
                break

    if found is not None:
        intersections.remove(found)
        duplicates.append(found)
    intersections.append(intersection)


def check_unused(intersection, duplicates, intersections):
    """Check if a "valid" ``intersection`` is already in ``intersections``.

    This assumes that

    * ``intersection`` will have at least one of ``s == 0.0`` or ``t == 0.0``
    * At least one of the intersections in ``intersections`` is classified as
      ``COINCIDENT_UNUSED``.

    Args:
        intersection (.Intersection): An intersection to be added.
        duplicates (List[.Intersection]): List of duplicate intersections.
        intersections (List[.Intersection]): List of "accepted" (i.e.
            non-duplicate) intersections.

    Returns:
        bool: Indicates if the ``intersection`` is a duplicate.
    """
    for other in intersections:
        if (
            other.interior_curve == UNUSED_T
            and intersection.index_first == other.index_first
            and intersection.index_second == other.index_second
        ):
            if intersection.s == 0.0 and other.s == 0.0:
                duplicates.append(intersection)
                return True

            if intersection.t == 0.0 and other.t == 0.0:
                duplicates.append(intersection)
                return True

    return False


def add_intersection(  # pylint: disable=too-many-arguments
    index1,
    s,
    index2,
    t,
    interior_curve,
    edge_nodes1,
    edge_nodes2,
    duplicates,
    intersections,
):
    """Create an :class:`Intersection` and append.

    The intersection will be classified as either a duplicate or a valid
    intersection and appended to one of ``duplicates`` or ``intersections``
    depending on that classification.

    Args:
        index1 (int): The index (among 0, 1, 2) of the first edge in the
            intersection.
        s (float): The parameter along the first curve of the intersection.
        index2 (int): The index (among 0, 1, 2) of the second edge in the
            intersection.
        t (float): The parameter along the second curve of the intersection.
        interior_curve (Optional[.IntersectionClassification]): The
            classification of the intersection, if known. If :data:`None`,
            the classification will be computed below.
        edge_nodes1 (Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]): The
            nodes of the three edges of the first triangle being intersected.
        edge_nodes2 (Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]): The
            nodes of the three edges of the second triangle being intersected.
        duplicates (List[.Intersection]): List of duplicate intersections.
        intersections (List[.Intersection]): List of "accepted" (i.e.
            non-duplicate) intersections.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-locals
    edge_end, is_corner, intersection_args = _py_triangle_helpers.handle_ends(
        index1, s, index2, t
    )
    if edge_end:
        intersection = _py_intersection_helpers.Intersection(
            *intersection_args
        )
        intersection.interior_curve = interior_curve
        if interior_curve == UNUSED_T:
            add_edge_end_unused(intersection, duplicates, intersections)
        else:
            duplicates.append(intersection)
    else:
        intersection = _py_intersection_helpers.Intersection(
            index1, s, index2, t
        )
        if is_corner:
            is_duplicate = check_unused(
                intersection, duplicates, intersections
            )
            if is_duplicate:
                return

        # Classify the intersection.
        if interior_curve is None:
            interior_curve = _py_triangle_helpers.classify_intersection(
                intersection, edge_nodes1, edge_nodes2
            )
        intersection.interior_curve = interior_curve
        intersections.append(intersection)


def classify_coincident(st_vals, coincident):
    r"""Determine if coincident parameters are "unused".

    .. note::

       This is a helper for :func:`triangle_intersections`.

    In the case that ``coincident`` is :data:`True`, then we'll have two
    sets of parameters :math:`(s_1, t_1)` and :math:`(s_2, t_2)`.

    If one of :math:`s1 < s2` or :math:`t1 < t2` is not satisfied, the
    coincident segments will be moving in opposite directions, hence don't
    define an interior of an intersection.

    .. warning::

       In the "coincident" case, this assumes, but doesn't check, that
       ``st_vals`` is ``2 x 2``.

    Args:
        st_vals (numpy.ndarray): ``2 X N`` array of intersection parameters.
        coincident (bool): Flag indicating if the intersections are the
            endpoints of coincident segments of two curves.

    Returns:
        Optional[.IntersectionClassification]: The classification of the
        intersections.
    """
    if not coincident:
        return None

    if st_vals[0, 0] >= st_vals[0, 1] or st_vals[1, 0] >= st_vals[1, 1]:
        return UNUSED_T

    else:
        return CLASSIFICATION_T.COINCIDENT


def should_use(intersection):
    """Check if an intersection can be used as part of a curved polygon.

    Will return :data:`True` if the intersection is classified as
    :attr:`~.IntersectionClassification.FIRST`,
    :attr:`~.IntersectionClassification.SECOND` or
    :attr:`~.IntersectionClassification.COINCIDENT` or if the intersection
    is classified is a corner / edge end which is classified as
    :attr:`~.IntersectionClassification.TANGENT_FIRST` or
    :attr:`~.IntersectionClassification.TANGENT_SECOND`.

    Args:
        intersection (.Intersection): An intersection to be added.

    Returns:
        bool: Indicating if the intersection will be used.
    """
    if intersection.interior_curve in ACCEPTABLE_CLASSIFICATIONS:
        return True

    if intersection.interior_curve in TANGENT_CLASSIFICATIONS:
        return intersection.s == 0.0 or intersection.t == 0.0

    return False


def triangle_intersections(edge_nodes1, edge_nodes2, all_intersections):
    """Find all intersections among edges of two triangles.

    This treats intersections which have ``s == 1.0`` or ``t == 1.0``
    as duplicates. The duplicates may be checked by the caller, e.g.
    by :func:`verify_duplicates`.

    Args:
        edge_nodes1 (Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]): The
            nodes of the three edges of the first triangle being intersected.
        edge_nodes2 (Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]): The
            nodes of the three edges of the second triangle being intersected.
        all_intersections (Callable): A helper that intersects B |eacute| zier
            curves. Takes the nodes of each curve as input and returns an
            array (``2 x N``) of intersections.

    Returns:
        Tuple[list, list, list, set]: 4-tuple of

        * The actual "unique" :class:`Intersection`-s
        * Duplicate :class:`Intersection`-s encountered (these will be
          corner intersections)
        * Intersections that won't be used, such as a tangent intersection
          along an edge
        * All the intersection classifications encountered
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-locals
    intersections = []
    duplicates = []
    for index1, nodes1 in enumerate(edge_nodes1):
        for index2, nodes2 in enumerate(edge_nodes2):
            st_vals, coincident = all_intersections(nodes1, nodes2)
            interior_curve = classify_coincident(st_vals, coincident)
            for s, t in st_vals.T:
                add_intersection(
                    index1,
                    s,
                    index2,
                    t,
                    interior_curve,
                    edge_nodes1,
                    edge_nodes2,
                    duplicates,
                    intersections,
                )
    all_types = set()
    to_keep = []
    unused = []
    for intersection in intersections:
        all_types.add(intersection.interior_curve)
        # Only keep the intersections which are "acceptable".
        if should_use(intersection):
            to_keep.append(intersection)
        else:
            unused.append(intersection)
    return to_keep, duplicates, unused, all_types


def generic_intersect(
    nodes1, degree1, nodes2, degree2, verify, all_intersections
):
    r"""Find all intersections among edges of two triangles.

    This treats intersections which have ``s == 1.0`` or ``t == 1.0``
    as duplicates. The duplicates will be checked by :func:`verify_duplicates`
    if ``verify`` is :data:`True`.

    Args:
        nodes1 (numpy.ndarray): The nodes defining the first triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree1 (int): The degree of the triangle given by ``nodes1``.
        nodes2 (numpy.ndarray): The nodes defining the second triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree2 (int): The degree of the triangle given by ``nodes2``.
        verify (Optional[bool]): Indicates if duplicate intersections
            should be checked.
        all_intersections (Callable): A helper that intersects B |eacute| zier
            curves. Takes the nodes of each curve as input and returns an
            array (``2 x N``) of intersections.

    Returns:
        Tuple[Optional[list], Optional[bool], tuple]: 3-tuple of

        * List of "edge info" lists. Each list represents a curved polygon
          and contains 3-tuples of edge index, start and end (see the
          output of :func:`ends_to_curve`).
        * "Contained" boolean. If not :data:`None`, indicates
          that one of the triangles is contained in the other.
        * The nodes of three edges of the first triangle being intersected
          followed by the nodes of the three edges of the second.
    """
    bbox_int = _py_geometric_intersection.bbox_intersect(nodes1, nodes2)
    if bbox_int != INTERSECTION_T:
        return [], None, ()

    # We need **all** edges (as nodes).
    edge_nodes1 = _py_triangle_helpers.compute_edge_nodes(nodes1, degree1)
    edge_nodes2 = _py_triangle_helpers.compute_edge_nodes(nodes2, degree2)
    # Run through **all** pairs of edges.
    intersections, duplicates, unused, all_types = triangle_intersections(
        edge_nodes1, edge_nodes2, all_intersections
    )
    edge_infos, contained = _py_triangle_helpers.combine_intersections(
        intersections, nodes1, degree1, nodes2, degree2, all_types
    )
    # Verify the duplicates and intersection if need be.
    if verify:
        verify_duplicates(duplicates, intersections + unused)
        verify_edge_segments(edge_infos)
    if edge_infos is None or edge_infos == []:
        return edge_infos, contained, ()

    return edge_infos, contained, edge_nodes1 + edge_nodes2


def geometric_intersect(nodes1, degree1, nodes2, degree2, verify):
    r"""Find all intersections among edges of two triangles.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Uses :func:`generic_intersect` with the
    :attr:`~.IntersectionStrategy.GEOMETRIC` intersection strategy.

    Args:
        nodes1 (numpy.ndarray): The nodes defining the first triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree1 (int): The degree of the triangle given by ``nodes1``.
        nodes2 (numpy.ndarray): The nodes defining the second triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree2 (int): The degree of the triangle given by ``nodes2``.
        verify (Optional[bool]): Indicates if duplicate intersections
            should be checked.

    Returns:
        Tuple[Optional[list], Optional[bool], tuple]: 3-tuple of

        * List of "edge info" lists. Each list represents a curved polygon
          and contains 3-tuples of edge index, start and end (see the
          output of :func:`ends_to_curve`).
        * "Contained" boolean. If not :data:`None`, indicates
          that one of the triangles is contained in the other.
        * The nodes of three edges of the first triangle being intersected
          followed by the nodes of the three edges of the second.
    """
    all_intersections = _py_geometric_intersection.all_intersections
    return generic_intersect(
        nodes1, degree1, nodes2, degree2, verify, all_intersections
    )


def algebraic_intersect(nodes1, degree1, nodes2, degree2, verify):
    r"""Find all intersections among edges of two triangles.

    Uses :func:`generic_intersect` with the
    :attr:`~.IntersectionStrategy.ALGEBRAIC` intersection strategy.

    Args:
        nodes1 (numpy.ndarray): The nodes defining the first triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree1 (int): The degree of the triangle given by ``nodes1``.
        nodes2 (numpy.ndarray): The nodes defining the second triangle in
            the intersection (assumed in :math:\mathbf{R}^2`).
        degree2 (int): The degree of the triangle given by ``nodes2``.
        verify (Optional[bool]): Indicates if duplicate intersections
            should be checked.

    Returns:
        Tuple[Optional[list], Optional[bool], tuple]: 3-tuple of

        * List of "edge info" lists. Each list represents a curved polygon
          and contains 3-tuples of edge index, start and end (see the
          output of :func:`ends_to_curve`).
        * "Contained" boolean. If not :data:`None`, indicates
          that one of the triangles is contained in the other.
        * The nodes of three edges of the first triangle being intersected
          followed by the nodes of the three edges of the second.
    """
    all_intersections = _algebraic_intersection.all_intersections
    return generic_intersect(
        nodes1, degree1, nodes2, degree2, verify, all_intersections
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup file for bezier."""

import os
import sys

import pkg_resources
import setuptools

import setup_helpers
import setup_helpers_macos
import setup_helpers_windows


VERSION = "2020.1.14"  # Also in ``codemeta.json`` and ``__init__.py``.
AUTHOR = "Danny Hermes"  # Also in ``__init__.py``.
README_FILENAME = os.path.join(os.path.dirname(__file__), "README.rst")
NUMPY_MESSAGE = """\
Error: NumPy needs to be installed first. It can be installed via:

$ python    -m pip install numpy
$ python3.8 -m pip install numpy
$ # OR
$ conda install numpy
"""
MISSING_F90_MESSAGE = """\
No Fortran 90 compiler found.

Skipping Fortran speedups via binary extension module.
"""
NO_EXTENSION_ENV = "BEZIER_NO_EXTENSION"
NO_SPEEDUPS_MESSAGE = """\
The {} environment variable has been used to explicitly disable the
building of the binary extension module.
""".format(
    NO_EXTENSION_ENV
)
REQUIREMENTS = ("numpy >= 1.18.1",)
EXTRAS_REQUIRE = {"full": ["scipy >= 1.4.1", "sympy >= 1.5.1"]}
DESCRIPTION = (
    u"Helper for B\u00e9zier Curves, Triangles, and Higher Order Objects"
)


def is_installed(requirement):
    try:
        pkg_resources.require(requirement)
    except pkg_resources.ResolutionError:
        return False

    else:
        return True


def require_numpy():
    if not is_installed("numpy>=1.9.0"):
        print(NUMPY_MESSAGE, file=sys.stderr)
        sys.exit(1)


def extension_modules():
    if NO_EXTENSION_ENV in os.environ:
        print(NO_SPEEDUPS_MESSAGE, file=sys.stderr)
        return []

    require_numpy()
    if setup_helpers.BuildFortranThenExt.has_f90_compiler():
        return setup_helpers.extension_modules()

    else:
        print(MISSING_F90_MESSAGE, file=sys.stderr)
        return []


def make_readme():
    with open(README_FILENAME, "r") as file_obj:
        return file_obj.read()


def setup():
    setuptools.setup(
        name="bezier",
        version=VERSION,
        description=DESCRIPTION,
        author=AUTHOR,
        author_email="daniel.j.hermes@gmail.com",
        long_description=make_readme(),
        scripts=(),
        url="https://github.com/dhermes/bezier",
        project_urls={
            "Documentation": "https://bezier.readthedocs.io/",
            "Changelog": (
                "https://bezier.readthedocs.io/en/2020.1.14/releases/index.html"
            ),
            "Issue Tracker": "https://github.com/dhermes/bezier/issues",
        },
        keywords=["Geometry", "Curve", "Bezier", "Intersection", "Python"],
        packages=["bezier"],
        package_dir={"": os.path.join("src", "python")},
        license="Apache 2.0",
        platforms="Posix; macOS; Windows",
        package_data={
            "bezier": [
                "*.pxd",
                os.path.join("include", "*.h"),
                os.path.join("include", "bezier", "*.h"),
                os.path.join("lib", "*.a"),
                os.path.join("lib", "*.lib"),
                os.path.join("extra-dll", "*.dll"),
            ]
        },
        zip_safe=True,
        install_requires=REQUIREMENTS,
        extras_require=EXTRAS_REQUIRE,
        ext_modules=extension_modules(),
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Mathematics",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
        ],
        cmdclass={"build_ext": setup_helpers.BuildFortranThenExt},
    )


def main():
    # Add any "patches" needed for the Fortran compiler.
    setup_helpers.BuildFortranThenExt.PATCH_FUNCTIONS[:] = [
        setup_helpers.patch_f90_compiler,
        setup_helpers_macos.patch_f90_compiler,
        setup_helpers_windows.patch_f90_compiler,
    ]
    setup_helpers_windows.patch_cmd(setup_helpers.BuildFortranThenExt)
    setup()


if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Legacy members that have been replaced or renamed."""

import warnings

from bezier import triangle


DEPRECATION_MSG = (
    "`bezier.Surface` is deprecated and has been replaced by "
    "`bezier.Triangle`. It will be removed in a future release."
)


class Surface(triangle.Triangle):
    """Legacy alias for B |eacute| zier Surface class.

    This will be deprecated in a future release.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        super(Surface, self).__init__(*args, **kwargs)
